(module disagreement
	(include "dctp.sch")
	(export (disagreement-linking-step link subgoal rel-vector) (decompose-disequation disequation-cell disequation-cdr)
		(cell-of-subgoal subgoal) (delete-disagreement-set) (insert-disagreement-set subgoals disagreement-set)
		(make-eq-disagreement-set upper-linked-lit lower-linked-lit lower-clause selected-side)
		(make-disagreement-set upper-linked-lit lower-linked-lit lower-clause)
		(make-decomposition-set disequation-lit))
	(import (globals "globals.scm") (unification "unification.scm") (display "display.scm") (dctp "dctp.scm")
		(normal-form "normal-form.scm") (max-sg-number link "link.scm") (unrenamed-termlist-copy rewrite "rewrite.scm")
		(pruning "pruning.scm") (insert-literal parser "parser.scm") (index "index.scm")
		(current-subgoal-weight select "select.scm") (subsumption "subsumption.scm")))


;;;*************************************************************************
;; ------------------------------------------------------------
;; Disagreement linking
;; ------------------------------------------------------------

(define (disagreement-linking-step link subgoal rel-vector)
  (cond
   ((eq? (link-misc link) 'normal)     (symmectric-linking-step link subgoal))
   ((eq? (link-misc link) 'transposed) (symmectric-linking-step link subgoal))
   ((not (link-litb link))         (decomposition-step link subgoal))
   ((eq-subgoal? (link-litb link)) (eq-disagreement-step link subgoal))
   (else                           (disagreement-step link subgoal))))
      
(define (symmectric-linking-step link subgoal)
  (let ((clausea (subgoal-clause (link-lita link)))
	(clauseb (subgoal-clause (link-litb link)))
	(depth (subgoal-depth subgoal))
	(copya #f) (copyb #f) copylita transposed-terms
	(new-subgoals #f))
;     (display "symmetric entry: ")
;     (display-subgoal (link-lita link))
;     (display " and " )
;     (display-subgoal (link-litb link))
;     (display " mode ")
;     (display (link-misc link))
;     (newline)
    (if (eq? (link-misc link) 'transposed)
	(set! transposed-terms (list (cadr (subgoal-terms (link-lita link))) (car (subgoal-terms (link-lita link))))))
    (if (or (and (eq? (link-misc link) 'transposed) (unify-lists-stop transposed-terms (subgoal-terms (link-litb link))))
	    (unify-lists-stop (subgoal-terms (link-lita link)) (subgoal-terms (link-litb link))))
	(begin
	  (set! total-inferences (+ total-inferences 1))
	  (if dewey-start
	      (if (>= total-inferences dewey-start)
		  (begin
		    (set! dewey-numbers #t)
		    (set! dewey-start #f))))
	  (if dewey-numbers
	      (begin
		(if (eq? (link-misc link) 'transposed)
		    (display "transposed "))
		(display "linking-step with link depths ") 
		(display (subgoal-depth (link-lita link)))
		(display " and ") 
		(display (subgoal-depth (link-litb link)))
		(display " below subgoal ")
		(display-subgoal subgoal)
		(newline)
		(print "+++++++++++++++++++++++++++++++++++++")
		(display-clauses (list clausea clauseb))
		(newline)
		(print "+++++++++++++++++++++++++++++++++++++")))
	  (if (or tautologies
		  (and (non-tautological clausea) (non-tautological clauseb)))
	      (let ((variant-a #f) (variant-b #f) (regular-a #f) (regular-b #f) (subsumed-a #f) (subsumed-b #f)
		    (keep-a #f) (keep-b #f)
		    (unsimplifieds-a (unsimplifiable-subgoals (clause-subgoals clausea))) 
		    (unsimplifieds-b (unsimplifiable-subgoals (clause-subgoals clauseb))))
		(if full-variants (set-linked-clauses-tested clausea clauseb))
		(set! variant-a (clause-index-variant? clausea))
		(if (not variant-a)
		    (begin
		      (set! subsumed-a (clause-subsumed clausea))
		      (if (not subsumed-a)
			  (begin
			    (set! regular-a  (ground-regular clausea))
			    (if (and regular-a full-variants)
				(begin
				  (compute-current-weights unsimplifieds-a)
				  (set! variant-a (clause-variant-subsumed unsimplifieds-a))))))))
		(set! variant-b (clause-index-variant? clauseb))
		(if (not variant-b)
		    (begin
		      (set! subsumed-b (clause-subsumed clauseb))
		      (if (not subsumed-b)
			  (begin
			    (set! regular-b  (ground-regular clauseb))
			    (if (and regular-b full-variants)
				(begin
				  (set! unsimplifieds-b  (unsimplifiable-subgoals (clause-subgoals clauseb)))
				  (compute-current-weights unsimplifieds-b)
				  (set! variant-b (clause-variant-subsumed unsimplifieds-b))))))))
		(if full-variants (unset-linked-clauses clausea clauseb))
		(if (and couples (non-unit-clause clausea) (non-unit-clause clauseb)
			 (not (eq? variant-a clausea)) (not (eq? variant-b clauseb))
			 (or variant-a variant-b subsumed-a subsumed-b (not regular-a) (not regular-b)))
		    (begin 
		      (set! coupled-eliminations (+ 1 coupled-eliminations))
		      (if dewey-numbers
			  (begin
			    (display "coupled elimination") 
			    (newline))))
		    (begin
		      (if (and (not (or subsumed-a variant-a)) regular-a)
			  (begin
			    (if (not unsimplifieds-a)
				(set! unsimplifieds-a  (unsimplifiable-subgoals (clause-subgoals clausea))))
			    (set! copya (renamed-and-splitted-copy clausea unsimplifieds-a 'upper 
								   (link-lita link) (+ depth 1)))
			    (set! depth (+ depth (length copya)))
			    (clauses-path-vector-set copya)
			    (set-clause-path (car copya) subgoal)
			    (set-subgoal-next subgoal (car copya))
			    (set! nr-simplifications (+ nr-simplifications (- (clause-unsimplified clausea)
									      (length unsimplifieds-a))))
			    (set! new-subgoals (clause-subgoals (car copya)))))
		      (if (and (not (or subsumed-b variant-b)) regular-b)
			  (begin
			    (if (not unsimplifieds-b)
				(set! unsimplifieds-b  (unsimplifiable-subgoals (clause-subgoals clauseb))))
			    (set! copyb (renamed-and-splitted-copy clauseb unsimplifieds-b 'lower 
								   (link-litb link) (+ depth 1)))
			    (set! depth (+ depth (length copyb)))
			    (clauses-path-vector-set copyb)
			    (set! nr-simplifications (+ nr-simplifications (- (clause-unsimplified clauseb)
									      (length unsimplifieds-b))))
			    (if (not new-subgoals)
				(begin
				  (set-clause-path (car copyb) subgoal)
				  (set-subgoal-next subgoal (car copyb))
				  (set! new-subgoals (clause-subgoals (car copyb))))
				(let ((copylita (subgoal-in-copy (link-lita link) (clause-subgoals (car copya)))))
				  (set-clause-path (car copyb) copylita)
				  (set-subgoal-next copylita (car copyb))))))))
		(if new-subgoals
		    (begin
		      (set! successful-inferences (+ successful-inferences 1))
		      (if copya
			  (set! tableau-clauses (+ tableau-clauses (length copya))))
		      (if copyb
			  (set! tableau-clauses (+ tableau-clauses (length copyb))))
		      (if (> depth (+ 1 (subgoal-depth subgoal)))
			  (begin
			    (if dewey-numbers
				(begin
				  (display "attached new clauses on depths ") 
				  (display (+ (subgoal-depth subgoal) 1)) (display " to ") (display depth)
				  (newline))))
			  (begin
			    (if dewey-numbers
				(begin
				  (display "attached new clause on depth ") 
				  (display depth)
				  (newline)))))
		      (if dewey-numbers
			  (begin
			    (display "new clauses: ")
			    (if copya 
				(display-clauses copya)
				(display-clauses copyb))
			    (if (and copya copyb)
				(begin
				  (display " and ")
				  (display-clauses copyb)))
			    (newline))))))
	      (set-link-active link 'delete)) ;; tautological links can be discarded
	  (unbind-vars)
; 	  (display "symmetric exit: ")
; 	  (display-subgoal (link-lita link))
; 	  (display " and " )
; 	  (display-subgoal (link-litb link))
; 	  (newline)
	  new-subgoals)
	#f)))

(define (decomposition-step link subgoal)
  (let (decomposition-set old-subgoals (clausea (subgoal-clause (link-lita link))) (new-subgoals #f))
;     (display "decomposition entry: ")
;     (display-subgoal (link-lita link))
;     (newline)
    (trail-stop)
    (set! old-subgoals (clause-subgoals clausea))
    (set! decomposition-set (make-decomposition-set (link-lita link)))
    (set-clause-subgoals clausea (remq (link-lita link) old-subgoals))
    (if (pair? decomposition-set) (set-clause-subgoals clausea (insert-disagreement-set (clause-subgoals clausea) decomposition-set)))
    ;;(set! disequation-cell (cell-of-subgoal (link-lita link)))
    ;;(set! disequation-cdr (cdr disequation-cell))
    ;;(decompose-disequation disequation-cell disequation-cdr)
    (set! total-decompositions (+ total-decompositions 1))
    (if dewey-numbers
	(begin
	  (display "decomposition-step with link depth ") 
	  (display (subgoal-depth (link-lita link)))
	  (display " below subgoal ")
	  (display-subgoal subgoal)
	  (newline)
	  (print "+++++++++++++++++++++++++++++++++++++")
	  (display-clause clausea)
	  (newline)
	  (print "+++++++++++++++++++++++++++++++++++++")))
    (if (or tautologies (and (non-tautological clausea)))
	(let ((variant-a #f) (regular-a #f) (subsumed-a #f)
	      (keep-a #f) (copya #f) (depth (subgoal-depth subgoal))
	      (unsimplifieds-a (unsimplifiable-subgoals (clause-subgoals clausea))))
	  (if full-variants (set-clause-tested clausea))
	  (set! variant-a (clause-index-variant? clausea))
	  (if (not variant-a)
	      (begin
		(set! subsumed-a (clause-subsumed clausea))
		(if (not subsumed-a)
		    (begin
		      (set! regular-a  (ground-regular clausea))
		      (if (and regular-a full-variants)
			  (begin
			    (compute-current-weights unsimplifieds-a)
			    (set! variant-a (clause-variant-subsumed unsimplifieds-a))))))))
	  (if full-variants (unset-clause-tested clausea))
	  (if (or variant-a subsumed-a (not regular-a))
	      (begin 
		(if dewey-numbers
		    (begin
		      (display "decomposition elimination") 
		      (newline))))
	      (begin
		(if (and (not (or subsumed-a variant-a)) regular-a)
		    (begin
		      (if (not unsimplifieds-a)
			  (set! unsimplifieds-a  (unsimplifiable-subgoals (clause-subgoals clausea))))
		      (set! copya (renamed-and-splitted-copy clausea unsimplifieds-a 'upper 
							     (link-lita link) (+ depth 1)))
		      (set! depth (+ depth (length copya)))
		      (clauses-path-vector-set copya)
		      (set-clause-path (car copya) subgoal)
		      (set-subgoal-next subgoal (car copya))
		      (set! nr-simplifications (+ nr-simplifications (- (clause-unsimplified clausea)
									(length unsimplifieds-a))))
		      (set! new-subgoals (clause-subgoals (car copya)))))))
	  (if new-subgoals
	      (begin
		(set! successful-decompositions (+ successful-decompositions 1))
		(if copya
		    (set! tableau-clauses (+ tableau-clauses (length copya))))
		(if (> depth (+ 1 (subgoal-depth subgoal)))
		    (begin
		      (if dewey-numbers
			  (begin
			    (display "attached new clauses on depths ") 
			    (display (+ (subgoal-depth subgoal) 1)) (display " to ") (display depth)
			    (newline))))
		    (begin
		      (if dewey-numbers
			  (begin
			    (display "attached new clause on depth ") 
			    (display depth)
			    (newline)))))
		(if dewey-numbers
		    (begin
		      (display "new clauses: ")
		      (if copya 
			  (display-clauses copya))
		      (newline))))))
	(set-link-active link 'delete)) ;; tautological links can be discarded
    (if (pair? decomposition-set) (delete-disagreement-set))
    (set-clause-subgoals clausea old-subgoals)
    (unbind-vars)
    ;;(set-car! disequation-cell (link-lita link))
    ;;(set-cdr! disequation-cell disequation-cdr)
    ;;(unbind-vars)
;     (display "decomposition exit: ")
;     (display-subgoal (link-lita link))
;     (newline)
    new-subgoals))

(define (cell-of-subgoal subgoal)
  (let ((clause (subgoal-clause subgoal)))
    (if clause
	(cell-in-subgoals subgoal (clause-subgoals clause))
	#f)))

(define (cell-in-subgoals subgoal subgoals)
  (if (pair? subgoals)
      (if (eq? subgoal (car subgoals))
	  subgoals
	  (cell-in-subgoals subgoal (cdr subgoals)))
      #f))

(define (make-decomposition-set disequation-lit)
  (disagreement-set (cdr (ref (car (subgoal-terms disequation-lit))))
		    (cdr (ref (cadr (subgoal-terms disequation-lit))))
		    (subgoal-clause disequation-lit) 
		    (max-sg-number (clause-subgoals (subgoal-clause disequation-lit)))))

(define (decompose-disequation disequation-cell disequation-cdr)
  (let ((left-terms (cdr (car (subgoal-terms (car disequation-cell)))))
	(right-terms (cdr (cadr (subgoal-terms (car disequation-cell)))))
	(subgoal (car disequation-cell))
	new-subgoals)
    (set! new-subgoals (decomposition-subgoals left-terms right-terms 
					      (car disequation-cell) disequation-cdr 
					      (max-sg-number (clause-subgoals (subgoal-clause subgoal)))))
    (set-car! disequation-cell (car new-subgoals))
    (set-cdr! disequation-cell (cdr new-subgoals))))

(define (decomposition-subgoals left-terms right-terms disequation disequation-cdr max-sg)
  (if (pair? left-terms)
      (if (or (and (variable (ref (car left-terms))) (not (occurs-term (ref (car left-terms)) (ref (car right-terms)))))
	      (and (variable (ref (car right-terms))) (not (occurs-term (ref (car right-terms)) (ref (car left-terms))))))
	  (begin
	    (if (variable (ref (car left-terms)))
		(bind (ref (car left-terms)) (ref (car right-terms)))
		(bind (ref (car right-terms)) (ref (car left-terms))))
	    (decomposition-subgoals (cdr left-terms) (cdr right-terms) disequation disequation-cdr max-sg))
	  (let ((new-sg (new-subgoal)))
	    (set-subgoal-pred new-sg equality-pred)
	    (set-subgoal-sign new-sg '-)
	    (set-subgoal-terms new-sg (list (ref (car left-terms)) (ref (car right-terms))))
	    (set-subgoal-clause new-sg (subgoal-clause disequation))
	    (set-subgoal-number new-sg (+ 1 max-sg))
	    (cons new-sg (decomposition-subgoals (cdr left-terms) (cdr right-terms) disequation disequation-cdr (+ 1 max-sg)))))
      disequation-cdr))
	
(define (disagreement-step link subgoal)
  ;; this is the code for a disagreement linking step between non-eq literals. This means
  ;; that the linked literal is decomposed to get the disagreement set
  ;; we proceed as follows:
  ;; 1) find out which direction is indicated by (link-misc link). If it is 'upper, then link-lita gets replaced
  ;;    by the negation of link-litb, the disagreement set is formed and attached to the clause of link-lita.
  ;;    If link-misc is 'lower, the same happens to link-litb.
  ;;    forced instantiation is applied to both clauses before renaming. this requires particular care as both 
  ;;    clauses need to be copied before any meddling in the first place.
  ;;    More precisely:
  ;;    a) form the disagreement set of the linked literals
  ;;    b) form a rewrite-copy of the termlist of link-litb (in the 'upper case). i.e. a term copy with identical variables.
  ;;    c) save the termlist of link-lita somewhere and set the termlist of link-lita to the new copy. Insert the disagreement
  ;;       set. similar to the decomposition case but more complicated, as the disagreement literals cannot simply replace
  ;;       some other literal but must be inserted in an appropriate place in such a way that they can again be removed.
  ;;    d) apply forced instantiation
  ;;    e) produce renamed copies of both clauses
  ;;    f) check the renamed copies for redundancy (can we have earlier redundancy checks as well? forced instantiation
  ;;       applied as normal instantiation to the original variables and later retracted?)
  ;;    g) bring the linked clauses back to their initial state
  ;;    h) return the new clauses
;   (display "disagreement entry: ")
;   (display-subgoal (link-lita link))
;   (display " and " )
;   (display-subgoal (link-litb link))
;   (display " mode ")
;   (display (link-misc link))
;   (newline)
  (let ((depth (subgoal-depth subgoal))
	(copya #f) (copyb #f) copylita copylitb disagreement-set lower-linked-terms saved-lower-lits
	upper-linked-lit upper-clause lower-linked-lit lower-clause
	(new-subgoals #f))
    (if (eq? (link-misc link) 'upper)
	(begin
	  (set! upper-linked-lit (link-litb link))
	  (set! upper-clause (subgoal-clause upper-linked-lit))
	  (set! lower-linked-lit (link-lita link))
	  (set! lower-clause (subgoal-clause lower-linked-lit)))
	(begin
	  (set! upper-linked-lit (link-lita link))
	  (set! upper-clause (subgoal-clause upper-linked-lit))
	  (set! lower-linked-lit (link-litb link))
	  (set! lower-clause (subgoal-clause lower-linked-lit))))
    (trail-stop)
    (set! disagreement-set (make-disagreement-set upper-linked-lit lower-linked-lit lower-clause))
    (set! lower-linked-terms (subgoal-terms lower-linked-lit))
    (set-subgoal-terms lower-linked-lit (unrenamed-termlist-copy (subgoal-terms upper-linked-lit)))
    (set! saved-lower-lits (clause-subgoals lower-clause))
    ;; insert-disagreement-set must store the insertion position in global vars for delete-disagreement-set
    (if (pair? disagreement-set)
	(set-clause-subgoals lower-clause (insert-disagreement-set (clause-subgoals lower-clause) disagreement-set)))
    ;; that has already been done by make-disagreement-literals on the fly
    ;; (trailed-forced-instantiate disagreement-set)
    (if dewey-numbers
	(begin
	  (display "disagreement-linking-step with link depths ") 
	  (display (subgoal-depth (link-lita link)))
	  (display " and ") 
	  (display (subgoal-depth (link-litb link)))
	  (display " below subgoal ")
	  (display-subgoal subgoal)
	  (newline)
	  (print "+++++++++++++++++++++++++++++++++++++")
 	  (display-clause upper-clause)
	  (newline)
 	  (display-clause lower-clause)
	  (newline)
	  (print "+++++++++++++++++++++++++++++++++++++")))
    (set! total-dis-inferences (+ total-dis-inferences 1))
    (if (or tautologies
	    (and (non-tautological upper-clause) (non-tautological lower-clause)))
	(let ((variant-a #f) (variant-b #f) (regular-a #f) (regular-b #f) (subsumed-a #f) (subsumed-b #f) 
	      (unsimplifieds-a #f) (unsimplifieds-b #f))
	  (if full-variants (set-linked-clauses-tested upper-clause lower-clause))
	  (set! variant-a (clause-index-variant? upper-clause))
	  (if (not variant-a)
	      (begin
		(set! subsumed-a (clause-subsumed upper-clause))
		(if (not subsumed-a)
		    (begin
		      (set! regular-a  (ground-regular upper-clause))
		      (if (and regular-a full-variants)
			  (begin
			    (set! unsimplifieds-a  (unsimplifiable-subgoals (clause-subgoals upper-clause)))
			    (compute-current-weights unsimplifieds-a)
			    (set! variant-a (clause-variant-subsumed unsimplifieds-a))))))))
	  (set! variant-b (clause-index-variant? lower-clause))
	  (if (not variant-b)
	      (begin
		(set! subsumed-b (clause-subsumed lower-clause))
		(if (not subsumed-b)
		    (begin
		      (set! regular-b  (ground-regular lower-clause))
		      (if (and regular-b full-variants)
			  (begin
			    (set! unsimplifieds-b  (unsimplifiable-subgoals (clause-subgoals lower-clause)))
			    (compute-current-weights unsimplifieds-b)
			    (set! variant-b (clause-variant-subsumed unsimplifieds-b))))))))
	  (if full-variants (unset-linked-clauses upper-clause lower-clause))
	  (if (and (not (or subsumed-a variant-a)) regular-a)
	      (begin
		(if (not unsimplifieds-a)
		    (set! unsimplifieds-a  (unsimplifiable-subgoals (clause-subgoals upper-clause))))
		(set! copya (renamed-and-splitted-copy upper-clause unsimplifieds-a 'upper 
						       (link-lita link) (+ depth 1)))
		(set! depth (+ depth (length copya)))
		(clauses-path-vector-set copya)
		(set-clause-path (car copya) subgoal)
		(set-subgoal-next subgoal (car copya))
		(set! nr-simplifications (+ nr-simplifications (- (clause-unsimplified upper-clause)
								  (length unsimplifieds-a))))
		(set! new-subgoals (clause-subgoals (car copya)))))
	  (if (and (not (or subsumed-b variant-b)) regular-b)
	      (begin
		(if (not unsimplifieds-b)
		    (set! unsimplifieds-b  (unsimplifiable-subgoals (clause-subgoals lower-clause))))
		(set! copyb (renamed-and-splitted-copy lower-clause unsimplifieds-b 'lower 
						       (link-litb link) (+ depth 1)))
		(set! copylitb (subgoal-in-copy (link-litb link) (clause-subgoals (car copyb))))
		(set! depth (+ depth (length copyb)))
		(clauses-path-vector-set copyb)
		(set! nr-simplifications (+ nr-simplifications (+ 1 (- (clause-unsimplified lower-clause)
								       (length unsimplifieds-b)))))
		(if (not new-subgoals)
		    (begin
		      (set-clause-path (car copyb) subgoal)
		      (set-subgoal-next subgoal (car copyb))
		      (set! new-subgoals (clause-subgoals (car copyb))))
		    (let ((copylita (subgoal-in-copy (link-lita link) (clause-subgoals (car copya)))))
		      (set-clause-path (car copyb) copylita)
		      (set-subgoal-next copylita (car copyb))))))
	  (if new-subgoals
	      (begin
		(set! successful-dis-inferences (+ successful-dis-inferences 1))
		(if copya
		    (set! tableau-clauses (+ tableau-clauses (length copya))))
		(if copyb
		    (set! tableau-clauses (+ tableau-clauses (length copyb))))
		(if (> depth (+ 1 (subgoal-depth subgoal)))
		    (begin
		      (if dewey-numbers
			  (begin
			    (display "attached new clauses on depths ") 
			    (display (+ (subgoal-depth subgoal) 1)) (display " to ") (display depth)
			    (newline))))
		    (begin
		      (if dewey-numbers
			  (begin
			    (display "attached new clause on depth ") 
			    (display depth)
			    (newline)))))
		(if dewey-numbers
		    (begin
		      (display "new clauses: ")
		      (if copya 
			  (display-clauses copya)
			  (display-clauses copyb))
		      (if (and copya copyb)
			  (begin
			    (display " and ")
			    (display-clauses copyb)))
		      (newline))))))
	(set-link-active link 'delete)) ;; tautological links can be discarded
    ;; undo the effects of the disagreement step for the original clauses
    (if (pair? disagreement-set)
	(delete-disagreement-set))
    (set-clause-subgoals lower-clause saved-lower-lits)
    (set-subgoal-terms lower-linked-lit lower-linked-terms)
    (unbind-vars)
;     (display "disagreement exit: ")
;     (display-subgoal (link-lita link))
;     (display " and " )
;     (display-subgoal (link-litb link))
;     (newline)
    new-subgoals))

(define (make-disagreement-set upper-linked-lit lower-linked-lit lower-clause)
  (disagreement-set (subgoal-terms upper-linked-lit) 
		    (subgoal-terms lower-linked-lit) 
		    lower-clause 
		    (max-sg-number (clause-subgoals lower-clause))))

(define (disagreement-set upper-terms lower-terms lower-clause max-sg)
  (if (pair? upper-terms)
      (let ((upper (ref (car upper-terms))) (lower (ref (car lower-terms))))
	(if (eq? upper lower)
	    (disagreement-set (cdr upper-terms) (cdr lower-terms) lower-clause max-sg)
	    (if (and (variable upper) (not (occurs-term upper lower)))
		(begin
		  ;;(display "binding ")
		  ;;(display-ref upper)
		  ;;(display "(")
		  ;;(display (car upper))
		  ;;(display ")")
		  (bind upper lower)
		  ;;(display " to ")
		  ;;(display-ref (ref upper))
		  ;;(newline)
		  (disagreement-set (cdr upper-terms) (cdr lower-terms) lower-clause max-sg))
		(if (and (variable lower) (not (occurs-term lower upper)))
		    (begin
		      ;;(display "binding ")
		      ;;(display-ref lower)
		      ;;(display "(")
		      ;;(display (car lower))
		      ;;(display ")")
		      (bind lower upper)
		      ;;(display " to ")
		      ;;(display-ref (ref lower))
		      ;;(newline)
		      (disagreement-set (cdr upper-terms) (cdr lower-terms) lower-clause max-sg))
		    (let ((new-sg (new-subgoal)))
		      (set-subgoal-pred new-sg equality-pred)
		      (set-subgoal-sign new-sg '-)
		      (set-subgoal-terms new-sg (list (ref (car upper-terms)) (ref (car lower-terms))))
		      (set-subgoal-clause new-sg lower-clause)
		      (set-subgoal-number new-sg (+ 1 max-sg))
		      (set-subgoal-weight new-sg (current-subgoal-weight new-sg))
		      (cons new-sg (disagreement-set (cdr upper-terms) (cdr lower-terms) lower-clause (+ 1 max-sg))))))))
      '()))

(define saved-lits-cell #f)
(define saved-lits-cdr '())

(define (insert-disagreement-set subgoals disagreement-set)
  (if (pair? subgoals)
      (if (sign-predsym< (car disagreement-set) (car subgoals))
	  ;; special-case: if the first disagreement is smaller than the current subgoal, the current subgoal
	  ;; must be the first subgoal of the clause
	  (begin
	    (set! saved-lits-cell #f)
	    (set! saved-lits-cdr subgoals)
	    (append! disagreement-set subgoals))
	  (if (or (null? (cdr subgoals)) (sign-predsym< (car disagreement-set) (cadr subgoals)))
	      (begin
		(set! saved-lits-cell subgoals)
		(set! saved-lits-cdr (cdr subgoals))
		(set-cdr! subgoals (append! disagreement-set (cdr subgoals)))
		subgoals)
	      (begin
		(insert-disagreement-set (cdr subgoals) disagreement-set)
		subgoals)))
      disagreement-set))
	  
(define (delete-disagreement-set)
  (if saved-lits-cell
      (begin
	(set-cdr! saved-lits-cell saved-lits-cdr)
	(set! saved-lits-cell #f))))

(define (eq-disagreement-step link subgoal)
  ;; this is the code for a disagreement linking step between eq literals. This means
  ;; that on term of the disequation is decomposed to get the disagreement set (modulo the other sides)
  ;; we proceed as follows:
  ;; 1) find out which linked literal is the disequation. If it is link-lita, then link-litb gets replaced
  ;;    by the negation of link-lita, the disagreement set is formed and attached to the clause of link-litb.
  ;;    If link-lita is the equation, the same happens to link-lita.
  ;;    find out which equation side is indicated for decomposition by (link-misc link). if it is 'left, then 
  ;;    the subterms of the left side form are decomposed, if it is 'right, then the other way round. for the
  ;;    disequation, always the left term is used, as the disequation has been reordered in solve-subgoal.
  ;;    forced instantiation is applied to both clauses before renaming. this requires particular care as both 
  ;;    clauses need to be copied before any meddling in the first place.
  ;;    More precisely:
  ;;    a) form the disagreement set of the linked literals
  ;;    b) form a rewrite-copy of the termlist of link-litb (in the 'upper case). i.e. a term copy with identical variables.
  ;;    c) save the termlist of link-lita somewhere and set the termlist of link-lita to the new copy. Insert the disagreement
  ;;       set. similar to the decomposition case but more complicated, as the disagreement literals cannot simply replace
  ;;       some other literal but must be inserted in an appropriate place in such a way that they can again be removed.
  ;;    d) apply forced instantiation
  ;;    e) produce renamed copies of both clauses
  ;;    f) check the renamed copies for redundancy (can we have earlier redundancy checks as well? forced instantiation
  ;;       applied as normal instantiation to the original variables and later retracted?)
  ;;    g) bring the linked clauses back to their initial state
  ;;    h) return the new clauses
;   (display "eq-disagreement entry: ")
;   (display-subgoal (link-lita link))
;   (display " and " )
;   (display-subgoal (link-litb link))
;   (display " mode ")
;   (display (link-misc link))
;   (newline)
  (let ((depth (subgoal-depth subgoal))
	(copya #f) (copyb #f) copylita copylitb disagreement-set lower-linked-terms saved-lower-lits
	upper-linked-lit upper-clause lower-linked-lit lower-clause
	(new-subgoals #f))
    (if (equation? (link-lita link))
	(begin
	  (set! upper-linked-lit (link-litb link))
	  (set! upper-clause (subgoal-clause upper-linked-lit))
	  (set! lower-linked-lit (link-lita link))
	  (set! lower-clause (subgoal-clause lower-linked-lit)))
	(begin
	  (set! upper-linked-lit (link-lita link))
	  (set! upper-clause (subgoal-clause upper-linked-lit))
	  (set! lower-linked-lit (link-litb link))
	  (set! lower-clause (subgoal-clause lower-linked-lit))))
    ;;(trail-stop)
    (if (eq? (link-misc link) 'left)
	(top-level-unify-stop (car (subgoal-terms upper-linked-lit)) (car (subgoal-terms lower-linked-lit)))
	(top-level-unify-stop (car (subgoal-terms upper-linked-lit)) (cadr (subgoal-terms lower-linked-lit))))
    (set! disagreement-set (make-eq-disagreement-set upper-linked-lit lower-linked-lit lower-clause (link-misc link)))
    (set! lower-linked-terms (subgoal-terms lower-linked-lit))
    (set-subgoal-terms lower-linked-lit (unrenamed-termlist-copy (subgoal-terms upper-linked-lit)))
    (set! saved-lower-lits (clause-subgoals lower-clause))
    ;; insert-disagreement-set must store the insertion position in global vars for delete-disagreement-set
    (if (pair? disagreement-set)
	(set-clause-subgoals lower-clause (insert-disagreement-set (clause-subgoals lower-clause) disagreement-set)))
    ;; that has already been done by make-disagreement-literals on the fly
    ;; (trailed-forced-instantiate disagreement-set)
    (if dewey-numbers
	(begin
	  (display "eq-disagreement-linking-step with link depths ") 
	  (display (subgoal-depth (link-lita link)))
	  (display " and ") 
	  (display (subgoal-depth (link-litb link)))
	  (display " below subgoal ")
	  (display-subgoal subgoal)
	  (if (eq? (link-misc link) 'left)
	      (display " into left side")
	      (display " into right side"))
	  (newline)
	  (print "+++++++++++++++++++++++++++++++++++++")
 	  (display-clause upper-clause)
	  (newline)
 	  (display-clause lower-clause)
	  (newline)
	  (print "+++++++++++++++++++++++++++++++++++++")))
    (set! total-dis-inferences (+ total-dis-inferences 1))
    (if (or tautologies
	    (and (non-tautological upper-clause) (non-tautological lower-clause)))
	(let ((variant-a #f) (variant-b #f) (regular-a #f) (regular-b #f) (subsumed-a #f) (subsumed-b #f) 
	      (unsimplifieds-a #f) (unsimplifieds-b #f))
	  (if full-variants (set-linked-clauses-tested upper-clause lower-clause))
	  (set! variant-a (clause-index-variant? upper-clause))
	  (if (not variant-a)
	      (begin
		(set! subsumed-a (clause-subsumed upper-clause))
		(if (not subsumed-a)
		    (begin
		      (set! regular-a  (ground-regular upper-clause))
		      (if (and regular-a full-variants)
			  (begin
			    (set! unsimplifieds-a  (unsimplifiable-subgoals (clause-subgoals upper-clause)))
			    (compute-current-weights unsimplifieds-a)
			    (set! variant-a (clause-variant-subsumed unsimplifieds-a))))))))
	  (set! variant-b (clause-index-variant? lower-clause))
	  (if (not variant-b)
	      (begin
		(set! subsumed-b (clause-subsumed lower-clause))
		(if (not subsumed-b)
		    (begin
		      (set! regular-b  (ground-regular lower-clause))
		      (if (and regular-b full-variants)
			  (begin
			    (set! unsimplifieds-b  (unsimplifiable-subgoals (clause-subgoals lower-clause)))
			    (compute-current-weights unsimplifieds-b)
			    (set! variant-b (clause-variant-subsumed unsimplifieds-b))))))))
	  (if full-variants (unset-linked-clauses upper-clause lower-clause))
	  (if (and (not (or subsumed-a variant-a)) regular-a)
	      (begin
		(if (not unsimplifieds-a)
		    (set! unsimplifieds-a  (unsimplifiable-subgoals (clause-subgoals upper-clause))))
		(set! copya (renamed-and-splitted-copy upper-clause unsimplifieds-a 'upper 
						       (link-lita link) (+ depth 1)))
		(set! depth (+ depth (length copya)))
		(clauses-path-vector-set copya)
		(set-clause-path (car copya) subgoal)
		(set-subgoal-next subgoal (car copya))
		(set! nr-simplifications (+ nr-simplifications (- (clause-unsimplified upper-clause)
								  (length unsimplifieds-a))))
		(set! new-subgoals (clause-subgoals (car copya)))))
	  (if (and (not (or subsumed-b variant-b)) regular-b)
	      (begin
		(if (not unsimplifieds-b)
		    (set! unsimplifieds-b  (unsimplifiable-subgoals (clause-subgoals lower-clause))))
		(set! copyb (renamed-and-splitted-copy lower-clause unsimplifieds-b 'lower 
						       (link-litb link) (+ depth 1)))
		(set! copylitb (subgoal-in-copy (link-litb link) (clause-subgoals (car copyb))))
		(set! depth (+ depth (length copyb)))
		(clauses-path-vector-set copyb)
		(set! nr-simplifications (+ nr-simplifications (+ 1 (- (clause-unsimplified lower-clause)
								       (length unsimplifieds-b)))))
		(if (not new-subgoals)
		    (begin
		      (set-clause-path (car copyb) subgoal)
		      (set-subgoal-next subgoal (car copyb))
		      (set! new-subgoals (clause-subgoals (car copyb))))
		    (let ((copylita (subgoal-in-copy (link-lita link) (clause-subgoals (car copya)))))
		      (set-clause-path (car copyb) copylita)
		      (set-subgoal-next copylita (car copyb))))))
	  (if new-subgoals
	      (begin
		(set! successful-dis-inferences (+ successful-dis-inferences 1))
		(if copya
		    (set! tableau-clauses (+ tableau-clauses (length copya))))
		(if copyb
		    (set! tableau-clauses (+ tableau-clauses (length copyb))))
		(if (> depth (+ 1 (subgoal-depth subgoal)))
		    (begin
		      (if dewey-numbers
			  (begin
			    (display "attached new clauses on depths ") 
			    (display (+ (subgoal-depth subgoal) 1)) (display " to ") (display depth)
			    (newline))))
		    (begin
		      (if dewey-numbers
			  (begin
			    (display "attached new clause on depth ") 
			    (display depth)
			    (newline)))))
		(if dewey-numbers
		    (begin
		      (display "new clauses: ")
		      (if copya 
			  (display-clauses copya)
			  (display-clauses copyb))
		      (if (and copya copyb)
			  (begin
			    (display " and ")
			    (display-clauses copyb)))
		      (newline))))))
	(set-link-active link 'delete)) ;; tautological links can be discarded
    ;; undo the effects of the disagreement step for the original clauses
    (if (pair? disagreement-set)
	(delete-disagreement-set))
    (set-clause-subgoals lower-clause saved-lower-lits)
    (set-subgoal-terms lower-linked-lit lower-linked-terms)
    (unbind-vars)
;     (display "eq-disagreement exit: ")
;     (display-subgoal (link-lita link))
;     (display " and " )
;     (display-subgoal (link-litb link))
;     (newline)
    new-subgoals))

(define (make-eq-disagreement-set upper-linked-lit lower-linked-lit lower-clause selected-side)
  (let (sub-dis other-side-dis)
    (if (eq? selected-side 'left)
	(begin
	  ;;(print "check a")
	  ;;(display-ref-list (subgoal-terms upper-linked-lit))
	  ;;(newline)
	  ;;(display-ref-list (subgoal-terms lower-linked-lit))
	  ;;(newline)
	  (set! sub-dis (disagreement-set (cdr (ref (car (subgoal-terms upper-linked-lit))))
					  (cdr (ref (car (subgoal-terms lower-linked-lit))))
					  lower-clause 
					  (max-sg-number (clause-subgoals lower-clause))))
	  ;;(display-subgoal-list sub-dis)
	  ;;(newline)
	  ;;(print "check b")
	  (if (pair? sub-dis)
	      (set! other-side-dis (disagreement-set (cdr (subgoal-terms upper-linked-lit))
						     (cdr (subgoal-terms lower-linked-lit))
						     lower-clause
						     (max-sg-number sub-dis)))
	      (set! other-side-dis (disagreement-set (cdr (subgoal-terms upper-linked-lit))
						     (cdr (subgoal-terms lower-linked-lit))
						     lower-clause
						     (max-sg-number (clause-subgoals lower-clause)))))
	  (append! sub-dis other-side-dis))
	(begin
	  ;;(print "check c," selected-side)
	  ;;(display-ref-list (subgoal-terms upper-linked-lit))
	  ;;(newline)
	  ;;(display-ref-list (subgoal-terms lower-linked-lit))
	  ;;(newline)
	  (set! sub-dis (disagreement-set (cdr (ref (car (subgoal-terms upper-linked-lit))))
					  (cdr (ref (cadr (subgoal-terms lower-linked-lit))))
					  lower-clause 
					  (max-sg-number (clause-subgoals lower-clause))))
	  ;;(print "check d")
	  (if (pair? sub-dis)
	      (set! other-side-dis (disagreement-set (cdr (subgoal-terms upper-linked-lit))
						     (list (ref (car (subgoal-terms lower-linked-lit))))
						     lower-clause
						     (max-sg-number sub-dis)))
	      (set! other-side-dis (disagreement-set (cdr (subgoal-terms upper-linked-lit))
						     (list (ref (car (subgoal-terms lower-linked-lit))))
						     lower-clause
						     (max-sg-number (clause-subgoals lower-clause)))))
	  (append! sub-dis other-side-dis)))))
