(module initial-path
	(include "dctp.sch")
	(export (select-initial-path clauses) (initial-active-path clauses predsg) (make-initial-link-list subgoal)
		(clauses-check-orientation clauses) (add-reflexivity-axiom) (clauses-orientate-disequations clauses) 
		(subgoals-orientate-disequations subgoals))
	(import (globals "globals.scm") (index "index.scm") (unification "unification.scm") (link "link.scm") (kbo "kbo.scm")
		(display "display.scm") (select "select.scm") (path "path.scm") (pruning "pruning.scm") 
		(clause-subsumed dctp "dctp.scm") (ground-closed dctp "dctp.scm")))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Variables global to this module
;; ------------------------------------------------------------

(define all-clauses-selected #f)
(define one-new-clause-selected #f)
(define one-new-subgoal-selected #f)

;;;*************************************************************************
;; ------------------------------------------------------------
;; Initial path functions
;; ------------------------------------------------------------

(define (initial-active-path clauses predsg)
  (if (pair? clauses)
      (let (sg)
	(set! initial-length (+ initial-length 1))
	(if (clause-relevant (car clauses))
	    (set-relevant-predicate-weights (clause-subgoals (car clauses))))
	(compute-subgoal-weights (clause-subgoals (car clauses)))
	;;(set! sg (get-selected-subgoal (clause-subgoals (car clauses))))
	(if (and preprocess least-initial-links (not equality))
	    (set! sg (select-least-linked-subgoal (clause-subgoals (car clauses))))
	    (set! sg (select-subgoal (clause-subgoals (car clauses)) #f)))
	(reset-subgoal-links (clause-subgoals (car clauses)))
	(if (not sg)
	    (begin
	       (print "empty input clause deduced by unit simplification!")
	       (set! sg (car (clause-subgoals (car clauses))))))
	(unmark-subgoals (clause-subgoals (car clauses)))
	(if (and (not preprocess) (unit-clause (car clauses)))
	    (insert-unit-in-indexes sg))
 	(set-clause-path (car clauses) predsg)
	(if (and (or disagreement (not ordered-eq)) (disequation? sg))
	    (begin
	      (if (unit-subgoal sg)
		  (delete-unit-from-indexes sg))
	      (check-orientation sg)
	      (if (unit-subgoal sg)
		  (insert-unit-in-indexes sg))))
	(if (and (equation? sg) (not disagreement)) (add-path-literal sg))
	(if predsg
	    (begin
	      (set-subgoal-next predsg (car clauses))
	      (set-clause-depth (car clauses) (+ 1 (subgoal-depth predsg))))
	    (begin
	      (set! input-root sg)
	      (set-clause-depth (car clauses) (- (length clauses)))))
	(if dewey-numbers
	    (begin
	      (display-subgoal sg)
	      (display " ")
	      (display-clause (subgoal-clause sg))
	      (newline)))
	(if (pair? (cdr clauses))
	    (initial-active-path (cdr clauses) sg)
	    sg))
      (begin
	(set! initial-select #f)
	#f)))

(define (set-relevant-predicate-weights subgoals)
  (if (pair? subgoals)
      (begin
	(set-predsym-weight (subgoal-pred (car subgoals)) 0)
	(set-relevant-termlist-weights (subgoal-terms (car subgoals)))
	(set-relevant-predicate-weights (cdr subgoals)))))
  
(define (set-relevant-termlist-weights terms)
  (if (pair? terms)
      (begin
	(set-relevant-term-weights (car terms))
	(set-relevant-termlist-weights (cdr terms)))))

(define (set-relevant-term-weights term)
  (if (variable term)
      (set! relevant-terms-ground #f)
      (if (constant term)
          (set-functor-weight (car term) 0)
	  (begin
	    (set-functor-weight (car term) 1)
	    (set-relevant-termlist-weights (cdr term))))))

(define (reset-subgoal-links subgoals)
  (if (pair? subgoals)
      (begin
	(set-subgoal-links (car subgoals) '())
	(reset-subgoal-links (cdr subgoals)))))

(define (make-initial-link-list subgoal)
  ;; the adding of path equations has been moved to initial-active-path so that all positive equalities
  ;; are available for eq-linking in case that is required
  (if subgoal
      (if (or (not (make-initial-link-list (subgoal-predecessor subgoal)))
	      (or (and (unit-subgoal subgoal) (subgoal-simplifiable subgoal))
		  (and (unit-subgoal subgoal) (ground-closed subgoal (subgoal-predecessor subgoal) (new-rel-vector 0)))))
	  #f
	  (begin
	    (if (not (subgoal-unlinkable subgoal))
		(begin
		  (if (> debug-level 4)
		      (begin
			(display "making initial links for subgoal ")
			(display-subgoal subgoal)
			(display " on depth ")
			(display (subgoal-depth subgoal))
			(newline)))
		  (if (or disagreement (not (and equality (eq-subgoal? subgoal))))
		      (if (eq-subgoal? subgoal)
			  (if (disequation? subgoal)
			      (begin
				(insert-decomposition-link subgoal)
				(disequation-insert-links subgoal (potential-linked-lits subgoal)))
			      (equation-insert-links subgoal (potential-linked-lits subgoal)))
			  (insert-links subgoal (potential-linked-lits subgoal))))
		  ;; NOTA BENE: Sign selection does not preserve completeness. The condition below
		  ;; must be replaced by (if (and equality (not disagreement))...) only.
		  ;;(if (and equality (not disagreement) 
			   ;;(or ordered-eq (disequation? subgoal)
			       ;;(and (not (eq-subgoal? subgoal)) (eq? (subgoal-sign subgoal) selected-sign))))
		  (if (and equality (not disagreement) )
		      (insert-eq-links subgoal (positive-eq-lits)))
		  ;;(if (> debug-level 4)
		      ;;(begin
			;;(display "checking for initial factoring links")
			;;(newline)))
 		  ;;(if factoring (insert-factoring-links subgoal))
		  (if (and equality (not disagreement) (disequation? subgoal))
		      (begin
			(if (> debug-level 4)
			    (begin
			      (display "checking for initial reflexivity link")
			      (newline)))
			(insert-reflexivity-link subgoal)))))
	    (if (clause-relevant (subgoal-clause subgoal))
		(set! initial-relevant-links (+ initial-relevant-links (length (subgoal-links subgoal)))))
	    (set! initial-links (+ initial-links (length (subgoal-links subgoal))))
	    (if (or (not (equation? subgoal)) disagreement) (add-path-literal subgoal))
	    #t))
      #t))

(define (clause-selected clause)
  (subgoals-selected (clause-subgoals clause)))

(define (subgoals-selected subgoals)
  (if (pair? subgoals)
      (if (subgoal-selected (car subgoals))
	  #t
	  (subgoals-selected (cdr subgoals)))
      #f))

(define (select-least-linked-subgoal subgoals)
  (if (pair? subgoals)
      (let ((least-rest (select-least-linked-subgoal (cdr subgoals))))
	(if least-rest
	    (if (< (length (subgoal-links least-rest)) (length (subgoal-links (car subgoals))))
		(car subgoals)
		(if (= (length (subgoal-links least-rest)) (length (subgoal-links (car subgoals))))
		    (if (subgoal-> (car subgoals) least-rest)
			(car subgoals)
			least-rest)
		    least-rest))
	    (car subgoals)))
      #f))

(define (select-initial-path clauses)
  ;;(display "marking relevant clauses...")
  (mark-relevant-clauses clauses)
  ;;(display "done")
  ;;(newline)
  (do ()
      (all-clauses-selected #t)
    ;;(display "loop for selecting initial subgoals...")
    ;;(newline)
    (set! all-clauses-selected #t)
    (set! one-new-clause-selected #f)
    (set! one-new-subgoal-selected #f)
    (mark-subgoals-in-clauses clauses)
    ;;(display "done")
    ;;(newline)
    (if (and (not all-clauses-selected) (not one-new-subgoal-selected))
	(begin
	  (mark-unselected-clauses clauses)
	  (set! all-clauses-selected #t))))
  (reset-link-statistics))

(define (mark-relevant-clauses clauses)
  (if (pair? clauses)
      (if link-from-conjecture
	  (begin
	    (mark-relevant-clauses (cdr clauses))
	    (if (and (clause-relevant (car clauses)) (not (clause-selected (car clauses))))
		(mark-subgoal-in-clause (car clauses))))
	  (begin
	    (if (and (clause-relevant (car clauses)) (not (clause-selected (car clauses))))
		(mark-subgoal-in-clause (car clauses)))
	    (mark-relevant-clauses (cdr clauses))))
      #t))

(define (mark-unselected-clauses clauses)
  (if (pair? clauses)
      (if link-from-conjecture
	  (begin
	    (mark-unselected-clauses (cdr clauses))
	    (if (not (clause-selected (car clauses)))
		(begin
		  ;;(display-clause (car clauses))
		  ;;(newline)
		  ;;(display-links-of-subgoals (clause-subgoals (car clauses)))
		  (mark-subgoal-in-clause (car clauses)))))
	  (begin
	    (if (not (clause-selected (car clauses)))
		(begin
		  ;;(display-clause (car clauses))
		  ;;(newline)
		  ;;(display-links-of-subgoals (clause-subgoals (car clauses)))
		  (mark-subgoal-in-clause (car clauses))))
	    (mark-unselected-clauses (cdr clauses))))
      #t))

(define (mark-subgoals-in-clauses clauses)
  (if (pair? clauses)
      (if link-from-conjecture
	  (begin
	    (mark-subgoals-in-clauses (cdr clauses))
	    (if (clause-selected (car clauses))
		(mark-subgoal-in-clause (car clauses))
		(set! all-clauses-selected #f)))
	  (begin
	    (if (clause-selected (car clauses))
		(mark-subgoal-in-clause (car clauses))
		(set! all-clauses-selected #f))
	    (mark-subgoals-in-clauses (cdr clauses))))
      #t))

(define (mark-subgoal-in-clause clause)
  (let (sg)
    (if least-initial-links
	(set! sg (least-linked-subgoal (clause-subgoals clause)))
	(set! sg (widest-linked-subgoal (clause-subgoals clause))))
    (if sg
	(begin
	  (set! one-new-subgoal-selected #t)
	  (if (clause-selected clause) 
	      (set-subgoal-selected sg 'semi-selected)
	      (begin
		(set-subgoal-selected sg #t)))
	  (mark-linked-subgoals (best-of-links (subgoal-links sg) sg))))))

(define (widest-linked-subgoal subgoals)
  (if (null? subgoals)
      #f
      (if (subgoal-selected (car subgoals))
	  (widest-linked-subgoal (cdr subgoals))
	  (let ((sg (widest-linked-subgoal (cdr subgoals))))
	    (if (and sg (> (length (subgoal-links sg)) (length (subgoal-links (car subgoals)))))
		sg
		(if (subgoal-simplifiable (car subgoals)) #f (car subgoals)))))))

(define (least-linked-subgoal subgoals)
  (if (null? subgoals)
      #f
      (if (subgoal-selected (car subgoals))
	  (least-linked-subgoal (cdr subgoals))
	  (let ((sg (least-linked-subgoal (cdr subgoals))))
	    (if (and sg (<= (length (subgoal-links sg)) (length (subgoal-links (car subgoals)))))
		sg
		(if (subgoal-simplifiable (car subgoals)) #f (car subgoals)))))))

(define (best-of-links links subgoal)
  (if (null? links)
      '()
      (if (null? (cdr links))
	  links
	  (let (c1 c2
		(r (best-of-links (cdr links) subgoal)))
	    (set! c1 (compute-link-complexity subgoal (car r)))
	    (set! c2 (compute-link-complexity subgoal (car links)))
	    (if (eq? (subgoal-clause (car r)) (subgoal-clause (car links)))
		(if (> c1 c2)
		    (cons (car links) (cdr r))
		    r)
		(cons (car links) r))))))

(define (mark-following-links links)
  (if (pair? links)
      (begin
	(if (not (clause-selected (subgoal-clause (car links))))
	    (begin
	      (set-subgoal-selected (car links) #t)
	      (let (sg)
		(if least-initial-links
		    (set! sg (least-linked-subgoal (clause-subgoals (subgoal-clause (car links)))))
		    (set! sg (widest-linked-subgoal (clause-subgoals (subgoal-clause (car links))))))
		(if sg
		    (mark-following-links (best-of-links (subgoal-links sg) sg))))))
	(mark-following-links (cdr links)))))

(define (mark-linked-subgoals links)
  (if (pair? links)
      (begin
	(if (not (clause-selected (subgoal-clause (car links))))
	    (begin
	      (set-subgoal-selected (car links) #t)
	      ;;(print "marked clause " (input-number (clause-input (subgoal-clause (car links)))) "(" 
	      ;;(clause-depth (subgoal-clause (car links))) ")")
	      (set! one-new-clause-selected #t))
	(mark-linked-subgoals (cdr links))))))

(define (get-selected-subgoal subgoals)
  (if (null? subgoals)
      #f
      (if (eq? (subgoal-selected (car subgoals)) #t)
	  (car subgoals)
	  (get-selected-subgoal (cdr subgoals)))))

(define (unmark-subgoals subgoals)
  (if (pair? subgoals)
      (begin
	(set-subgoal-selected (car subgoals) #f)
	(unmark-subgoals (cdr subgoals)))))

(define (display-links-of-subgoals subgoals)
  (if (pair? subgoals)
      (begin
	(display-subgoal-list (subgoal-links (car subgoals)))
	(newline)
	(display-links-of-subgoals (cdr subgoals)))))


(define (clauses-check-orientation clauses)
  (if (pair? clauses)
      (begin
	(subgoals-check-orientation (clause-subgoals (car clauses)))
	(clauses-check-orientation (cdr clauses)))))

(define (subgoals-check-orientation subgoals)
  (if (pair? subgoals)
      (begin
	(if (eq-subgoal? (car subgoals))
	    (check-orientation (car subgoals)))
	(subgoals-check-orientation (cdr subgoals)))))

(define (clauses-orientate-disequations clauses)
  (if (pair? clauses)
      (begin
	(subgoals-orientate-disequations (clause-subgoals (car clauses)))
	(clauses-orientate-disequations (cdr clauses)))))

(define (subgoals-orientate-disequations subgoals)
  (if (pair? subgoals)
      (begin
	(if (eq-subgoal? (car subgoals))
	    (orientate-disequation (car subgoals)))
	(subgoals-check-orientation (cdr subgoals)))))

(define (add-reflexivity-axiom)
  (let ((new-sg (new-subgoal)) 
	(new-cl (new-clause))
	(new-in (new-input))
	(single-var (new-variable)))
    (set-variable-name single-var 'X)
    (set-subgoal-terms new-sg (list single-var single-var))
    (set-subgoal-sign new-sg '+)
    (set-subgoal-pred new-sg equality-pred)
    (set-subgoal-links new-sg '())
    (set-subgoal-number new-sg 1)
    (compute-subgoal-weights (list new-sg))
    (set-subgoal-simplified new-sg #f)
    (set-subgoal-unlinkable new-sg #t)
    (set-subgoal-clause new-sg new-cl)
    (set-clause-subgoals new-cl (list new-sg))
    (set-clause-nr-variables new-cl 1)
    (set-clause-variables new-cl (list single-var))
    (set-clause-input new-cl new-in)
    (set-clause-depth new-cl (- -1 (length (formula-clauses formula))))
    (set-clause-initial new-cl #t)
    (set-clause-unsimplified new-cl 1)
    (set-input-copies new-in (list new-cl))
    (set-input-length new-in 1)
    (set-input-type new-in 'axiom)
    (set-input-number new-in 0)
    (set-input-name new-in (string->symbol "reflexivity_axiom"))
    (insert-subgoal-in-gr-index new-sg)
    (insert-subgoal-in-index new-sg 'unit)
    new-sg))
