(module kbo
	(include "dctp.sch")
	(export kbo-precedences)
	(export (check-orientation subgoal) (compute-kbo-precedences functor-list) (kbo> s t) (orientate-disequation subgoal)
		(kbo-precedence term))
	(import (globals "globals.scm") (parser "parser.scm") (unification "unification.scm") (display "display.scm")
		(term-symbol-count select "select.scm") (ref-equal pruning "pruning.scm")))

(define functor-vector #f)
(define kbo-precedences (make-vector 1))

(define (check-orientation subgoal)
  (if (> debug-level 4)
      (begin
	(display "checking orientation of subgoal ")
	(display-subgoal subgoal)
	(newline)))
  (if (eq? (subgoal-pred subgoal) equality-pred)
      (let ((left (ref (car (subgoal-terms subgoal)))) (right (ref (cadr (subgoal-terms subgoal)))))
	(if (kbo> left right)
	    (begin
	      (set-subgoal-ordered-eq subgoal #t)
	      (if (> debug-level 4)
		  (begin
		    (display "subgoal ")
		    (display-subgoal subgoal)
		    (display " is kbo-orientated as is")
		    (newline))))
	    (if (kbo> right left)
		(begin
		  (set-subgoal-ordered-eq subgoal #t)
		  (set-subgoal-terms subgoal (list right left))
		  (if (> debug-level 4)
		      (begin
			(display "subgoal ")
			(display-subgoal subgoal)
			(display " is kbo-orientated with changed sides")
			(newline))))
		(set-subgoal-ordered-eq subgoal #f))))))

(define (orientate-disequation subgoal)
  (if (disequation? subgoal)
      (let ((left (ref (car (subgoal-terms subgoal)))) 
	    (right (ref (cadr (subgoal-terms subgoal)))))
	(if (variable left)
	    (set-subgoal-terms subgoal (list right left))
	    (if (not (variable right))
		(if (kbo> left right)
		    (set-subgoal-terms subgoal (list right left))
		    (if (not (kbo> right left))
			(let ((left-s (term-symbol-count left 1)) (right-s (term-symbol-count right 1)))
			  (if (> left-s right-s)
			      (set-subgoal-terms subgoal (list right left)))))))))))

;; FIXME: Definitiv unfertig, die letzten schliessende Klammern sind nur fuers Compilieren
(define (kbo-compare s t)
  (if (or (variable s) (variable t))
      (kbo-compare-vars s t)
      (let ((weight-s (kbo-weight s)) (weight-t (kbo-weight t)))
	(if (> weight-s weight-t)
	    (if (kbo-variable-occurrences> s t)
		'kbo-greater
		'kbo-uncomparable)
	    (if (> weight-t weight-s)
		(if (kbo-variable-occurrences> t s)
		    'kbo-lesser
		    'kbo-uncomparable))))))
		
		
  

(define (kbo> s t)
  (if (or (variable s) (variable t))
      (if (eq? 'kbo-greater (kbo-compare-vars s t))
	  #t
	  #f)
      (if (or (not (variable s)) (eq? s t))
	  (let ((weight-s (kbo-weight s)) (weight-t (kbo-weight t)))
	    (if (> weight-s weight-t)
		(kbo-variable-occurrences> s t)
		(if (= weight-s weight-t)
		    (if (> (kbo-precedence s) (kbo-precedence t))
			(kbo-variable-occurrences> s t)
			(if (and (= (kbo-precedence s) (kbo-precedence t))
				 (list-kbo> (cdr s) (cdr t)))
			    (kbo-variable-occurrences> s t)
			    #f))
		    #f)))
	  #f)))

(define (list-kbo> list-s list-t)
  (if (pair? list-s)
      (if (kbo> (ref (car list-s)) (ref (car list-t)))
	  #t
	  (if (ref-equal (ref (car list-s)) (ref (car list-t)))
	      (list-kbo> (cdr list-s) (cdr list-t))
	      #f))
      #f))

(define (kbo-weight term)
  (if (variable term)
      1
      (if (constant term)
	  1
	  (+ 1 (list-kbo-weight (cdr term))))))

(define (list-kbo-weight terms)
  (if (pair? terms)
      (+ (kbo-weight (ref (car terms))) (list-kbo-weight (cdr terms)))
      0))

(define (kbo-precedence term)
  (if (variable term)
      0
      (vector-ref kbo-precedences (- (functor-number (car term)) 1))))

(define (compute-kbo-precedences functor-list)
  (set! functor-vector (list->vector functor-list))
  ;;(set! functor-vector (sort functor-vector precedence<))
  (set! kbo-precedences (make-vector (vector-length functor-vector)))
  (do ((i 0 (+ i 1)))
      ((= i (vector-length functor-vector)) #t)
    ;;(display (functor-symbol (vector-ref functor-vector i)))
    ;;(display " ")
    ;;(display (functor-number (vector-ref functor-vector i)))
    ;;(newline)
    (vector-set! kbo-precedences (- (functor-number (vector-ref functor-vector i)) 1) (+ i 1))))

(define (precedence< f g)
  (if (< (functor-arity f) (functor-arity g))
      #t
      (if (= (functor-arity f) (functor-arity g))
	  (< (functor-number f) (functor-number g))
	  #f)))

(define (kbo-variable-occurrences> s t)
  (let ((tmp (var-occurrences-in-term (ref s) #t '())))
    (var-occurrences-in-term (ref t) #f tmp)))

(define (var-occurrences-in-termlist terms left vars)
  ;; finds all free variables in terms and inserts them into vars exactly once
  (if (not (null? terms))
      (let ((new-vars (var-occurrences-in-term (ref (car terms)) left vars)))
	(if new-vars
	    (var-occurrences-in-termlist (cdr terms) left new-vars)
	    #f))
      vars))

(define (var-occurrences-in-term term left vars)
  (if (free-variable term)
      (increment-or-insert-var term left vars)
      (if (not (constant term))
	  (var-occurrences-in-termlist (cdr term) left vars)
	  vars)))

(define (increment-or-insert-var term left vars)
  (if (pair? vars)
      (if (eq? term (vector-ref (car vars) 0)) ;; if the variable already occured in the termlist
	  (begin
	    (if left
		(vector-set! (car vars) 1 (+ (vector-ref (car vars) 1) 1))
		(vector-set! (car vars) 2 (+ (vector-ref (car vars) 2) 1)))
	    (if (> (vector-ref (car vars) 2) (vector-ref (car vars) 1))
		#f
		vars))
	  (if (pair? (cdr vars))
	      (if (increment-or-insert-var term left (cdr vars))
		  vars
		  #f)
	      (if left
		  (let ((new-var-entry (make-vector 3)))
		    (vector-set! new-var-entry 0 term)
		    (vector-set! new-var-entry 1 1)
		    (vector-set! new-var-entry 2 0)
		    (set-cdr! vars (list new-var-entry))
		    vars)
		  (begin
		    #f))))
      (if left
	  (let ((new-var-entry (make-vector 3)))
	    (vector-set! new-var-entry 0 term)
	    (vector-set! new-var-entry 1 1)
	    (vector-set! new-var-entry 2 0)
	    (list new-var-entry))
	  #f)))

(define (kbo-compare-vars s t)
  (if (variable s)
      (if (variable t)
	  (if (eq? s t)
	      `kbo-equal
	      `kbo-uncomparable)
	  (if (occurs-list s (cdr t))
	      `kbo-lesser
	      `kbo-uncomparable))
      (if (variable t)
	  (if (occurs-list t (cdr s))
	      `kbo-greater
	      `kbo-uncomparable)
	  (begin
	    (display "invalid kbo-var-comparison of terms")
	    (display-ref s)
	    (display " and ")
	    (display-ref t)
	    (newline)))))
	      
;;(define (kbo-compare-variable-occurrences s t))
