(module normal-form
	(include "dctp.sch")
	(export (renamed-subgoals subgoals) (non-equivalent-termlists k l) (renamed-termlist terms)
		(varlist-renamed? vars all-vars renamings) (unsimplifiable-subgoals subgoals)
		(non-equivalent-termlists-1 k l))
	(import (globals "globals.scm") (unification "unification.scm") (display "display.scm") (kbo "kbo.scm")
		(index "index.scm") (subgoal-in-list pruning "pruning.scm") (current-subgoal-weight select "select.scm")
		(rewrite-subgoal rewrite "rewrite.scm") (insert-literal parser "parser.scm")))


;;;*************************************************************************
;; ------------------------------------------------------------
;; Variables global to this module
;; ------------------------------------------------------------

(define current-class 1)

;;;*************************************************************************
;; ------------------------------------------------------------
;; Normal form functions
;; ------------------------------------------------------------

(define (non-equivalent-termlists k l)
  (let (result)
    (set! var-counter 0)
    (trail-stop)
    (set! result (non-equivalent-termlists-1 k l))
    (if (not result) (unbind-vars-and-undo-var-counter))
; 	(begin
; 	  (display "equivalent: ")
; 	  (display-ref-list k)
; 	  (display " and ")
; 	  (display-ref-list l)
; 	  (newline)
; 	  (unbind-vars-and-undo-var-counter)
; 	  (display "unbound:    ")
; 	  (display-subgoal-terms k)
; 	  (display " and ")
; 	  (display-subgoal-terms l)
; 	  (newline)))
    result))

(define (non-equivalent-termlists-1 k l)
  (if (pair? k)
      (let ((result (non-equivalent-terms (ref (car k)) (ref (car l)))))
	(if result
	    result
	    (non-equivalent-termlists-1 (cdr k) (cdr l))))
      #f))

(define (non-equivalent-terms s t)
  (if (eq? s t)
      #f
      (if (and (variable s) (variable t))
	  (begin
	    (bind s var-counter)
	    (bind t var-counter)
	    (set! var-counter (+ var-counter 2))
	    #f)
	  (if (and (non-variable s) (non-variable t) (eq? (car s) (car t)))
	      (non-equivalent-termlists-1 (cdr s) (cdr t))
	      (begin
		(unbind-vars-and-undo-var-counter)
		#t)))))

(define (non-equivalent-subgoals a b)
  (if (eq? (subgoal-sign a) (subgoal-sign b))
      (if (eq? (subgoal-pred a) (subgoal-pred b))
	  (non-equivalent-termlists (subgoal-terms a) (subgoal-terms b))
	  #t)
      #t))

(define (renamed-subgoals subgoals)
  (if (pair? subgoals)
      (let ((sg (new-subgoal)) (new-subgoals (renamed-subgoals (cdr subgoals))) same-sg (rewritten #f))
	(set! same-sg (subgoal-in-list (car subgoals) (cdr subgoals)))
	(if (and contract same-sg) 
	    (begin
	      (if (> debug-level 4)
		  (begin
		    (display "deleting multiple occurrence of subgoal ")
		    (display-subgoal (car subgoals))
		    (newline)))
	      (set! nr-contractions (+ nr-contractions 1))
	      (set-subgoal-misc (car subgoals) same-sg)
	      new-subgoals)
	    (begin
	      (set-subgoal-sign sg (subgoal-sign (car subgoals)))
	      (set-subgoal-pred sg (subgoal-pred (car subgoals)))
	      (set-subgoal-unlinkable sg (subgoal-unlinkable (car subgoals)))
	      (set-subgoal-selected sg #f)
	      (set-subgoal-simplified sg #f)
	      (set-subgoal-links sg '())
	      (set-subgoal-terms sg (renamed-termlist (subgoal-terms (car subgoals))))
	      (set-subgoal-number sg (subgoal-number (car subgoals)))
              ;;(if (and equality ordered-eq (or full-rewrite (and eq-rewrite (eq-subgoal? sg))))
	      ;;(set! rewritten (rewrite-subgoal sg)))
	      (if (and (not rewritten) (integer? (subgoal-misc (car subgoals))))
		  (set-subgoal-weight sg (subgoal-misc (car subgoals)))
		  (set-subgoal-weight sg (current-subgoal-weight sg)))
	      (set-subgoal-misc (car subgoals) sg)
	      (set-subgoal-ordered-eq sg (subgoal-ordered-eq (car subgoals)))
	      (if (or (and ordered-eq (eq-subgoal? sg) (or rewritten (not (subgoal-ordered-eq sg))))
		      (and (or (not ordered-eq) disagreement) (disequation? sg)))
		  (check-orientation sg))
	      ;;(cons sg new-subgoals))))
	      (insert-literal sg new-subgoals))))
      '()))

(define (renamed-termlist terms)
  (if (pair? terms)
      (if equality
	  (cons (renamed-term (ref (car terms))) (renamed-termlist (cdr terms)))
	  (cons (renamed-term (semi-ref (car terms))) (renamed-termlist (cdr terms))))
      '()))

(define (renamed-term term)
  (if (variable term)
      (if (variable-renaming term)
	  (variable-renaming term)
	  (let ((renaming (new-variable)))
	    (push-trail term)
	    (set-variable-renaming term renaming)
	    (set-variable-name renaming (variable-name term))
	    (if (bound-variable term)
		(set-variable-binding renaming (renamed-term (variable-binding term))))
	    renaming))
      (if (constant term)
	  term
	  (begin
	    (if (= (functor-arity (car term)) 0) (print "copying constant " term))
	    (cons (car term) (renamed-termlist (cdr term)))))))

; (define (semi-ref term)
; ;; if term is a <bound_variable> bound to another variable, its binding is returned, 
; ;; otherwise term is returned
;  (if (or (non-variable term) (free-variable term) (domain-element term)
; 	 (and (variable term) (non-variable (variable-binding term))))
;      term
;      (semi-ref (car term))))

(define (semi-ref term)
  (if (and (bound-variable term) (variable (variable-binding term)))
      (semi-ref (car term))
      term))

(define (unsimplifiable-subgoals subgoals)
  (if (pair? subgoals)
      (if (and simplify
	       (or (subgoal-simplified (car subgoals)) (match-subgoal-with-index (car subgoals))))
	  (begin
	    (if (and (> debug-level 4) (not (subgoal-simplified (car subgoals))))
		(begin
		  (display "simplified subgoal instance ")
		  (display-subgoal (car subgoals))
		  (newline)))
	    (set-subgoal-misc (car subgoals) #f)
	    (unsimplifiable-subgoals (cdr subgoals)))
	  (cons (car subgoals) (unsimplifiable-subgoals (cdr subgoals))))
      '()))     


;;;*************************************************************************
;; ------------------------------------------------------------
;; New check for variable renamings
;; ------------------------------------------------------------

(define (varlist-renamed? vars all-vars renamings)
  (if (pair? vars)
      (if (free-variable (car vars)) ;; if the current variable has not been bound at all
	  (varlist-renamed? (cdr vars) all-vars renamings)
	  (if (and (variable (ref (car vars))) ;; if the current variable has been bound to another variable
		   (not (memq (ref (car vars)) all-vars))  ;; that is not one from the var's own clause
		   (not (memq (ref (car vars)) renamings)));; or has been used to rename another var from the clause
	      (varlist-renamed? (cdr vars) all-vars (cons (ref (car vars)) renamings))
	      #f))
      (begin
	(set! total-renamings (+ total-renamings 1))
	#t)))
