(module preprocess
	(include "dctp.sch")
	(export nr-clauses nr-non-horn nr-units nr-literals nr-eq-literals nr-predicates nr-propositionals 
		nr-functors nr-constants nr-variables nr-singleton-variables nr-axioms nr-lemmas nr-hypotheses
		nr-conjectures  nr-all-negatives predicate-list nr-non-eq-literals)
	(export (preprocess-formula) (get-formula-statistics) (display-formula-statistics) (all-negative subgoals)
		(select-closure-constant))
	(import (globals "globals.scm") (link-count link "link.scm") (eq-link-count link "link.scm")
		(link-inc link "link.scm") (eq-link-inc link "link.scm") (delete-element-from-list link "link.scm")
		(index "index.scm") (eliminate-var-disequations dctp "dctp.scm") (dctp "dctp.scm") (kbo> kbo "kbo.scm")
		(check-orientation kbo "kbo.scm") (kbo-precedence kbo "kbo.scm") (connection "connection.scm")
		(display "display.scm")	(non-tautological pruning "pruning.scm") (subgoal-in-list pruning "pruning.scm")
		(parser "parser.scm") (renamed-subgoals normal-form "normal-form.scm") (splitting "splitting.scm")
		(subsumption "subsumption.scm") (unification "unification.scm") (kbo-precedences kbo "kbo.scm")))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Variables of global use
;; ------------------------------------------------------------

(define current-clause-variables '())
(define current-singleton-variables '())
(define predicate-list '())
(define propositional-list '())
(define functor-list '())
(define constant-list '())
(define current-positive-subgoals 0)
(define nr-clauses 0)               
(define nr-non-horn 0)              
(define nr-units 0)
(define nr-literals 0)              
(define nr-eq-literals 0)           
(define nr-non-eq-literals 0)           
(define max-clause-length 0)        
(define max-clause-variables 0)        
(define nr-predicates 0)            
(define nr-propositionals 0)        
(define max-pred-arity 0)           
(define min-pred-arity 0)           
(define nr-functors 0)              
(define nr-constants 0)             
(define max-func-arity 0)           
(define min-func-arity 0)           
(define nr-variables 0)             
(define nr-singleton-variables 0)   

(define nr-axioms 0)   
(define nr-lemmas 0)   
(define nr-hypotheses 0)   
(define nr-conjectures 0)   
(define nr-all-negatives 0)   

;;;*************************************************************************
;; ------------------------------------------------------------
;; Variables global to this module
;; ------------------------------------------------------------

(define formula-changed #f)
(define processing-conjecture #f)
(define clause-rewritten #f)
(define term-rewritten #f)

;;;*************************************************************************
;; ------------------------------------------------------------
;; Formula preprocessing functions
;; ------------------------------------------------------------

(define (preprocess-formula)
  (if (not silent) (print "starting preprocessing loop"))
  (set! formula-changed #f)
  (clear-clause-connections (formula-clauses formula))
  (if (clause-connections (formula-clauses formula))
      (begin
	;;(print "connection graph done")
	(if preprocess
	    (if (preprocess-clauses (formula-clauses formula))
		(begin
		  ;;(print "preprocessing loop done")
		  (if formula-changed
		      (preprocess-formula)
		      #t))
		#f)
	    #t))
      #f))

(define (preprocess-clauses clauses)
  (if (pair? clauses)
      (begin
	;;(display "preprocessing ")
	;;(display-clause  (car clauses))
	;;(newline)
	(if (preprocess-clause (car clauses))
	    (preprocess-clauses (cdr clauses))
	    #f))
      #t))

(define (preprocess-clause clause)
;   (display "preprocessing ")
;   (display-clause clause)
;   (newline)
  (if (not (or tautologies (non-tautological clause)))
      (begin
	(if dewey-numbers 
	    (begin 
	      (display "deleted tautological clause ")
	      (display-clause clause)
	      (newline)))
	(set! formula-changed #t)
	(remove-subgoals-from-links (clause-subgoals clause))
	(set-formula-clauses formula (delete-element-from-list clause (formula-clauses formula) (formula-clauses formula)))
	(if (unit-clause clause)
	    (delete-unit-from-indexes (car (clause-subgoals clause))))
	#t)
      (if (and (not equality) (pure-clause clause))
	  (begin
	    (if dewey-numbers 
		(begin 
		  (display "deleted pure clause ")
		  (display-clause clause)
		  (newline)))
	    (set! formula-changed #t)
	    (remove-subgoals-from-links (clause-subgoals clause))
	    (set-formula-clauses formula (delete-element-from-list clause (formula-clauses formula) (formula-clauses formula)))
	    (if (unit-clause clause)
		(delete-unit-from-indexes (car (clause-subgoals clause))))
	    #t)
	  (let ((subsumed (clause-subsumed clause)))
	    (if (and subsumption subsumed (not (eq? (subgoal-clause subsumed) clause)))
		(begin
		  (if dewey-numbers 
		      (begin 
			(display "deleted subsumed clause ")
			(display-clause clause)
			(newline)))
		  (set! formula-changed #t)
		  (remove-subgoals-from-links (clause-subgoals clause))
		  (set-formula-clauses formula (delete-element-from-list clause (formula-clauses formula) (formula-clauses formula)))
		  (if (unit-clause clause)
		      (delete-unit-from-indexes (car (clause-subgoals clause))))
		  #t)
		(let ((full-subsumed #f))
		  (if pre-subsume (set! full-subsumed (clause-full-subsumed-by-clauselist clause (formula-clauses formula))))
		  (if (and pre-subsume full-subsumed)
		      (begin
			(if dewey-numbers 
			    (begin 
			      (display "deleted clause ")
			      (display-clause clause)
			      (display " subsumed by ")
			      (display-clause full-subsumed)
			      (newline)))
			(set! formula-changed #t)
			(remove-subgoals-from-links (clause-subgoals clause))
			(set-formula-clauses formula (delete-element-from-list clause (formula-clauses formula) (formula-clauses formula)))
			(if (unit-clause clause)
			    (delete-unit-from-indexes (car (clause-subgoals clause))))
			#t)
		      (if (and demod (unit-clause clause) (equation? (car (clause-subgoals clause))) (is-demodulator clause))
			  (begin
			    ;; FIXME: The preprocessing needs to keep track of the number of
			    ;; equality literals in the input. if e.g. demodulation kills the only equation
			    ;; in the problem, equality can then be set to #f, purity can be applied etc.
			    (if (eq? (input-type (clause-input clause)) 'conjecture)
				(set! processing-conjecture #t))
			    (set-subgoal-ordered-eq (car (clause-subgoals clause)) #t)
			    (if (and (not (is-left-demodulator clause)) (is-right-demodulator clause))
				(set-subgoal-terms (car (clause-subgoals clause))
						   (list (cadr (subgoal-terms (car (clause-subgoals clause))))
							 (car (subgoal-terms (car (clause-subgoals clause)))))))
			    (if dewey-numbers 
				(begin 
				  (if (and (is-left-demodulator clause) (is-right-demodulator clause))
				      (display "applying symmectric demodulator ")
				      (display "applying demodulator "))
				  (display-clause clause)
				  (newline)))
			    (try-rewrite-clauses (car (clause-subgoals clause)) (formula-clauses formula))
			    (set! formula-changed #t)
			    (delete-unit-from-indexes (car (clause-subgoals clause)))
			    (set-subgoal-terms (car (clause-subgoals clause))
					       (list (cadr (subgoal-terms (car (clause-subgoals clause))))
						     (cadr (subgoal-terms (car (clause-subgoals clause))))))
			    (insert-unit-in-indexes (car (clause-subgoals clause)))
			    (set! processing-conjecture #f)
			    #t)
			  (if (and res-isol (not split) (not equality) (has-isolated-subgoal clause)
				   (or (not isol-limit) (<= (length (clause-subgoals clause)) isol-limit)))
			      (let ((isolated (has-isolated-subgoal clause)))
				(if (eq? (input-type (clause-input clause)) 'conjecture)
				    (set! processing-conjecture #t))
				(if dewey-numbers
				    (begin 
				      (display "found isolated subgoal ")
				      (display-subgoal isolated)
				      (display " in ")
				      (display-clause clause)
				      (newline)))
				(if (unit-clause clause)
				    (delete-unit-from-indexes isolated))
				(if (resolve-isolated-links isolated (subgoal-links isolated))
				    (begin
				      (remove-subgoals-from-links (remq! isolated (clause-subgoals clause)))
				      (set! formula-changed #t)
				      (set-formula-clauses formula (delete-element-from-list clause (formula-clauses formula) 
											     (formula-clauses formula)))
				      (set! processing-conjecture #f)
				      #t)
				    #f))
			      (let ((splitted '())(old-length (length (clause-subgoals clause))))
				(set-clause-subgoals clause (eliminate-var-disequations (clause-subgoals clause)))
				(set-clause-subgoals clause (delete-redundant-subgoals (clause-subgoals clause)))
				(set-clause-unsimplified clause (length (clause-subgoals clause)))
				(if (null? (clause-subgoals clause))
				    (begin
				      (if dewey-numbers
					  (begin 
					    (display "empty clause deduced by preprocessing")
					    (newline)))
				      (set-clause-subgoals clause (list falsum))
				      (set-clause-nr-variables clause 0)
				      (set-clause-variables clause '())
				      (set-clause-unsimplified clause 0)
				      #f)
				    (begin
				      (if (and isol split (not equality))
					  (let ((isolated (has-isolated-subgoal clause)))
					    (if (and isolated (not (propositional-subgoal? isolated)))
						(begin
						  (if dewey-numbers
						      (begin 
							(display "propagating substitutions for isolated subgoal ")
							(display-subgoal isolated)
							(display " in ")
							(display-clause clause)
							(newline)))
						  (propagate-isolated-unifiers isolated (subgoal-links isolated))))))
				      (set-clause-variables clause (variables-in-cleared-subgoals (clause-subgoals clause) '()))
				      (set-clause-nr-variables clause (length (clause-variables clause)))
				      (if (and rewrite-steps (unit-clause clause) (equation? (car (clause-subgoals clause))))
					  (begin
					    (if (> debug-level 5)
						(begin 
						  (display "rewriting with unit equation ")
						  (display-clause clause)
						  (newline)))
					    (try-rewrite-clauses (car (clause-subgoals clause)) (formula-clauses formula))))
				      (if (and (> old-length 1) (unit-clause clause))
					  (begin
					    (if dewey-numbers 
						(begin 
						  (display "generated new unit ")
						  (display-clause clause)
						  (newline)))
					    (insert-unit-in-indexes (car (clause-subgoals clause)))))
				      (if (and split (not (eq? (car (clause-subgoals clause)) falsum)))
					  (set! splitted (split-subgoals (clause-subgoals clause) '())))
				      (if (and split (> (length splitted) 1) (not (and (= (length splitted) 2) (one-list-is-ground splitted))))
					  (begin
					    (if (or dewey-numbers (> debug-level 4))
						(begin
						  (display "subgoals ")
						  (display-subgoal-list (clause-subgoals clause))
						  (display " can be split into ")
						  (display (length splitted))
						  (display " disjoint subsets")
						  (newline)))
					    (set! splitted (split-lists-to-clauses splitted (clause-depth clause) #f))
					    (set-clauses-input splitted (clause-input clause))
					    (set-clause-pointers-in-clauses splitted)
					    (if (or dewey-numbers (> debug-level 4))
						(begin
						  (display "new splitted clauses:")
						  (newline)
						  (display-clauses splitted)
						  (newline)))
					    (set! nr-splitted (+ nr-splitted (- (length splitted) 1)))
					    (set-formula-clauses formula (insert-splitted-clauses-in-clauselist 
									  splitted (remq! clause (formula-clauses formula))))
					    (set! formula-changed #t))
					  (if (not (eq? (length (clause-subgoals clause)) old-length))
					      (set! formula-changed #t)))
				      #t))))))))))))

(define (has-isolated-subgoal clause)
  (if equality
      #f
      (subgoals-isolated (clause-subgoals clause))))

(define (subgoals-isolated subgoals)
  (if (pair? subgoals)
      (if (and (all-links-singleton (subgoal-links (car subgoals))) (not (has-internal-links (car subgoals))))
	  (car subgoals)
	  (subgoals-isolated (cdr subgoals)))
      #f))

(define (all-links-singleton links)
  (if (pair? links)
      (if (or (> (length (subgoal-links (car links))) 1) (has-internal-links (car links)))
	  #f
	  (all-links-singleton (cdr links)))
      #t))
  
(define (propagate-isolated-unifiers subgoal links)
  (if (pair? links)
      (let ((other-clause (subgoal-clause (car links))) new-cl)
	(if (unit-clause other-clause) (delete-subgoal-from-index (car (clause-subgoals other-clause)) 'unit))
	(if (unify-lists-stop (subgoal-terms subgoal) (subgoal-terms (car links)))
	    (begin
	      (if (varlist-instantiated? (clause-variables other-clause))
		  (begin
		    (set! new-cl (renamed-copy other-clause (clause-subgoals other-clause)))
		    (set-clause-initial new-cl (clause-initial other-clause))
		    (set-clause-depth new-cl (clause-depth other-clause))
		    (add-links-for-renamed-subgoals (clause-subgoals other-clause))
		    (remove-subgoals-from-links (clause-subgoals other-clause))
		    (replace-clause-in-clauselist other-clause new-cl (formula-clauses formula))
		    (if (unit-clause new-cl) (insert-subgoal-in-index (car (clause-subgoals new-cl)) 'unit))
		    (if dewey-numbers
			(begin
			  (display "applied isolated unifier to ")
			  (display-clause new-cl)
			  (newline)))
		    (set! formula-changed #t))
		  (if (unit-clause other-clause) (insert-subgoal-in-index (car (clause-subgoals other-clause)) 'unit)))
	      (unbind-vars))
	    (if (unit-clause other-clause) (insert-subgoal-in-index (car (clause-subgoals other-clause)) 'unit)))
	(propagate-isolated-unifiers subgoal (cdr links)))))

(define (resolve-isolated-links subgoal links)
  (if (pair? links)
      (let ((other-clause (subgoal-clause (car links))) (new-cl #f) (falsified #f))
	(if (unit-clause other-clause)
	    (delete-unit-from-indexes (car (clause-subgoals other-clause))))
	(if (unify-lists-stop (subgoal-terms subgoal) (subgoal-terms (car links)))
	    (begin
	      (set! new-cl (new-clause))
	      (set-clause-foldup-lits new-cl '())
	      (set-clause-input new-cl (clause-input other-clause))
	      (set-clause-depth new-cl (clause-depth other-clause))
	      (set-clause-initial new-cl #t)
	      (trail-stop)
	      (set-clause-subgoals new-cl (renamed-subgoals (append (remq subgoal (clause-subgoals (subgoal-clause subgoal)))
								    (remq (car links) (clause-subgoals other-clause)))))
	      (unrename-vars)
	      (unbind-vars)
	      (set-clause-unsimplified new-cl (length (clause-subgoals new-cl)))
	      (if (null? (clause-subgoals new-cl))
		  (begin
		    (set! falsified #t)
		    (set-clause-subgoals new-cl (list falsum))))
	      (set-clause-variables new-cl (variables-in-subgoals (clause-subgoals new-cl) '()))
	      (set-clause-nr-variables new-cl (length (clause-variables new-cl)))
	      (clause-pointers-in-subgoals (clause-subgoals new-cl) new-cl)
	      (set-clause-initial new-cl (clause-initial other-clause))
	      (set-clause-depth new-cl (clause-depth other-clause))
	      (add-links-for-renamed-subgoals (remq (car links) (clause-subgoals other-clause)))
	      (add-links-for-renamed-resolving-subgoals (remq subgoal (clause-subgoals (subgoal-clause subgoal))))
	      ;;(if (eq? (input-type (clause-input (subgoal-clause subgoal))) 'conjecture)
	      (if processing-conjecture
		  (set-input-type (clause-input new-cl) 'conjecture))
	      (replace-clause-in-clauselist other-clause new-cl (formula-clauses formula))
	      (if (unit-clause new-cl)
		  (insert-unit-in-indexes (car (clause-subgoals new-cl))))
	      (if dewey-numbers
		  (begin
		    (display "resolved isolated connection to ")
		    (display-clause new-cl)
		    (newline))))
	    (begin
	      (if (unit-clause other-clause) (insert-unit-in-indexes (car (clause-subgoals other-clause))))))
	(if falsified
	    (begin
	      (if dewey-numbers
		  (begin
		    (display "resolving isolated connection produced empty clause")
		    (newline)))
	      #f)
	    (resolve-isolated-links subgoal (cdr links))))
      (begin
	#t)))

(define (add-links-for-renamed-resolving-subgoals subgoals)
  (if (pair? subgoals)
      (begin
	(add-links-for-renamed-resolving-copy (car subgoals) (subgoal-misc (car subgoals)) (subgoal-links (car subgoals)))
	(set-subgoal-misc (car subgoals) #f)
	(add-links-for-renamed-resolving-subgoals (cdr subgoals)))))

(define (add-links-for-renamed-subgoals subgoals)
  (if (pair? subgoals)
      (begin
	(add-links-for-renamed-copy (car subgoals) (subgoal-misc (car subgoals)) (subgoal-links (car subgoals)))
	(set-subgoal-misc (car subgoals) #f)
	(add-links-for-renamed-subgoals (cdr subgoals)))))

(define (add-links-for-renamed-resolving-copy subgoal copy-sg linked-subgoals)
   (if (pair? linked-subgoals)
      (begin
	 (if (unify-lists-stop (subgoal-terms copy-sg) (subgoal-terms (car linked-subgoals)))
	     (begin
	       (unbind-vars)
	       ;; Nota Bene: This must not be done at this point!
	       ;; The resolving clause can cause a number of isolated resolvents, and
	       ;; the original links are needed until all copies in all resolvents have
	       ;; gotten an analogous link!
	       ;;(replace-subgoal-in-links subgoal copy-sg (subgoal-links (car linked-subgoals)))
	       (set-subgoal-links (car linked-subgoals) (cons copy-sg (subgoal-links (car linked-subgoals))))
	       (set-subgoal-links copy-sg (cons (car linked-subgoals) (subgoal-links copy-sg)))))
	 (add-links-for-renamed-resolving-copy subgoal copy-sg (cdr linked-subgoals)))))

(define (add-links-for-renamed-copy subgoal copy-sg linked-subgoals)
   (if (pair? linked-subgoals)
      (begin
	 (if (unify-lists-stop (subgoal-terms copy-sg) (subgoal-terms (car linked-subgoals)))
	     (begin
	       (unbind-vars)
	       (replace-subgoal-in-links subgoal copy-sg (subgoal-links (car linked-subgoals)))
	       (set-subgoal-links copy-sg (cons (car linked-subgoals) (subgoal-links copy-sg))))
	     (begin
	       (set-subgoal-links (car linked-subgoals) (remq! subgoal (subgoal-links (car linked-subgoals))))))
	 (add-links-for-renamed-copy subgoal copy-sg (cdr linked-subgoals)))))

(define (remove-subgoal-from-links subgoal links)
  (if (pair? links)
      (begin
	 (set-subgoal-links (car links) (remq! subgoal (subgoal-links (car links))))
	 (remove-subgoal-from-links subgoal (cdr links)))))
	     
(define (remove-subgoals-from-links subgoals)	
  (if (pair? subgoals)
      (begin
	(remove-subgoal-from-links (car subgoals) (subgoal-links (car subgoals)))
	(remove-subgoals-from-links (cdr subgoals)))))

(define (replace-subgoal-in-linklists old-subgoal new-subgoal links)
  (if (pair? links)
      (begin
	(replace-subgoal-in-links old-subgoal new-subgoal (subgoal-links (car links)))
	(replace-subgoal-in-linklists old-subgoal new-subgoal (cdr links)))))

(define (replace-subgoal-in-links old-subgoal new-subgoal links)
  (if (pair? links)
      (if (eq? old-subgoal (car links))
	  (set-car! links new-subgoal)
	  (replace-subgoal-in-links old-subgoal new-subgoal (cdr links)))))
	
(define (replace-clause-in-clauselist old-clause new-cl clauses)
  (if (pair? clauses)
      (if (eq? old-clause (car clauses))
	  (set-car! clauses new-cl)
	  (replace-clause-in-clauselist old-clause new-cl (cdr clauses)))))

(define (varlist-instantiated? variables)
  (if (pair? variables)
      (if (free-variable (car variables))
	  (varlist-instantiated? (cdr variables))
	  #t)
      #f))

(define (has-internal-links subgoal)
  (let (rest-subgoals)
    (trail-stop)
    (set! rest-subgoals (renamed-subgoals (remq subgoal (clause-subgoals (subgoal-clause subgoal)))))
    (unrename-vars)
    (if (is-linked-to-subgoals subgoal rest-subgoals)
	#t
	#f)))

(define (is-linked-to-subgoals subgoal other-subgoals)
  (if (pair? other-subgoals)
      (if (and (not (eq? (subgoal-sign subgoal) (subgoal-sign (car other-subgoals))))
	       (eq? (subgoal-pred subgoal) (subgoal-pred (car other-subgoals)))
	       (unify-lists-stop (subgoal-terms subgoal) (subgoal-terms (car other-subgoals))))
	  (begin
	    (unbind-vars)
	    #t)
	  (is-linked-to-subgoals subgoal (cdr other-subgoals)))
      #f))

(define (pure-clause clause)
  (if (and (= (length (clause-subgoals clause)) 1) (eq? (car (clause-subgoals clause)) falsum))
      #f
      (pure-subgoals (clause-subgoals clause))))

(define (pure-subgoals subgoals)
  (if (pair? subgoals)
      (begin
	(if (= 0 (length (subgoal-links (car subgoals))))
	    (begin
	      (if dewey-numbers 
		  (begin 
		    (display "found pure subgoal ")
		    (display-subgoal (car subgoals))
		    (newline)))
	      (set! formula-changed #t)
	      (car subgoals))
	    (pure-subgoals (cdr subgoals))))
      #f))

(define (delete-redundant-subgoals subgoals)
  (if (pair? subgoals)
      (let ((matching-unit #f))
	(if simplify (set! matching-unit (match-subgoal-with-index (car subgoals))))
	(if (and simplify matching-unit)
	    (begin
	      (if dewey-numbers 
		  (begin 
		    (display "delete simplifiable subgoal ")
		    (display-subgoal (car subgoals))
		    (newline)))
	      (set! formula-changed #t)
	      (if (eq? (input-type (clause-input (subgoal-clause matching-unit))) 'conjecture)
		  (set-input-type (clause-input (subgoal-clause (car subgoals))) 'conjecture))
	      (remove-subgoal-from-links (car subgoals) (subgoal-links (car subgoals)))
	      (delete-redundant-subgoals (cdr subgoals)))
	    (let (same-sg (new-subgoals (delete-redundant-subgoals (cdr subgoals))))
	      (set! same-sg (subgoal-in-list (car subgoals) new-subgoals))
	      (if (and contract same-sg)
		  (begin
		    (if dewey-numbers
			(begin
			  (display "deleting multiple occurrence of subgoal ")
			  (display-subgoal (car subgoals))
			  (newline)))
		    (set! nr-contractions (+ nr-contractions 1))
		    (remove-subgoal-from-links (car subgoals) (subgoal-links (car subgoals)))
		    (set! formula-changed #t)
		    new-subgoals)
		  (cons (car subgoals) new-subgoals)))))
      '()))

(define (get-formula-statistics)
  (reset-formula-statistics)
  (get-clauses-statistics (formula-clauses formula))
  (if (= nr-clauses 0)
      (begin
	(if (not silent) (print "all clauses deleted by preprocessing")))
      (begin
	(if (= nr-all-negatives 0)
	    (if (not silent) (print "no all negative clause in formula")))
	(if (and (not norel) (= nr-clauses nr-conjectures))
	    (if (not silent) (print "all clauses of type conjecture ... restricting relevance to all-negative clauses")))
	(if (and (not norel) (= nr-conjectures 0))
	    (if (not silent) (print "no conjecture type clause ... restricting relevance to all-negative clauses")))
	(if (or norel (= nr-clauses nr-conjectures) (= nr-conjectures 0))
	    (set-all-negatives-relevant (formula-clauses formula)))
	(if (= nr-propositionals nr-predicates)
	    (begin
	      (if (not silent) (print "input formula is propositional"))
	      (set! groundable #t))
	    (if (= max-func-arity 0)
		(begin
		  (if (not silent) (print "input formula is in Bernays-Schnfinkel class"))
		  (set! groundable #t))))
	(if (= 0 nr-non-horn)
	    (begin
	      (if (not silent) (print "input formula is Horn"))
	      (set! horn #t)))
	(if (= nr-eq-literals 0)
	    (set! equality #f))
	(if (and equality (= nr-eq-literals nr-literals) (= nr-clauses nr-units))
	    (begin
	      (if (not silent) (print "input formula is in unit equality class"))
	      (set! unit-eq #t)
	      (set! full-variants #f)))))
  (set! link-inc nr-eq-literals)
  (set! eq-link-inc (* eq-bias nr-non-eq-literals))
  (set! link-count link-inc)
  (set! eq-link-count eq-link-inc))

(define (get-clauses-statistics clauses)
  (if (pair? clauses)
      (begin
	(cond ((eq? (input-type (clause-input (car clauses))) 'axiom)
	       (set! nr-axioms (+ nr-axioms 1)))
	      ((eq? (input-type (clause-input (car clauses))) 'lemma)
	       (set! nr-lemmas (+ nr-lemmas 1)))
	      ((eq? (input-type (clause-input (car clauses))) 'hypothesis)
	       (set! nr-hypotheses (+ nr-hypotheses 1)))
	      ((eq? (input-type (clause-input (car clauses))) 'conjecture)
	       (set! nr-conjectures (+ nr-conjectures 1))))
	(set! nr-clauses (+ nr-clauses 1))
	(if (unit-clause (car clauses))
	    (set! nr-units (+ nr-units 1)))
	(set! current-positive-subgoals 0)
	(if (> (length (clause-subgoals (car clauses))) max-clause-length)
	    (set! max-clause-length (length (clause-subgoals (car clauses)))))
	(set! nr-variables (+ nr-variables (length (clause-variables (car clauses)))))
	(set! current-clause-variables '())
	(set! current-singleton-variables '())
	(get-subgoals-statistics (clause-subgoals (car clauses)))
	(set! nr-singleton-variables (+ nr-singleton-variables (length current-singleton-variables)))
; 	(if (> (length current-singleton-variables) 0)
; 	    (begin
; 	      (display "singleton var ")
; 	      (display current-singleton-variables)
; 	      (display " found in ")
; 	      (display-clause (car clauses))
; 	      (newline)))
; 	(if (not (= (length current-clause-variables) (length (clause-variables (car clauses)))))
; 	    (print "mismatching clause lists!"))
	(if (> current-positive-subgoals 1)
	    (set! nr-non-horn (+ nr-non-horn 1)))
	(if (= current-positive-subgoals 0)
	    (set! nr-all-negatives (+ nr-all-negatives 1)))
	(get-clauses-statistics (cdr clauses)))))

(define (get-subgoals-statistics subgoals)
  (if (pair? subgoals)
      (begin
	(set! nr-literals (+ nr-literals 1))
	(if (eq? (subgoal-sign (car subgoals)) '+)
	    (set! current-positive-subgoals (+ current-positive-subgoals 1)))
	(if (eq? (subgoal-pred (car subgoals)) equality-pred)
	    (set! nr-eq-literals (+ nr-eq-literals 1))
	    (set! nr-non-eq-literals (+ nr-non-eq-literals 1)))
	(if (not (memq (subgoal-pred (car subgoals)) predicate-list))
	    (begin
	      (set! predicate-list (cons (subgoal-pred (car subgoals)) predicate-list))
	      (set! nr-predicates (+ nr-predicates 1))
	      ;;(set-predsym-number (subgoal-pred (car subgoals)) nr-predicates)
	      (if (= (predsym-arity (subgoal-pred (car subgoals))) 0)
		  (begin
		    (set! propositional-list (cons (subgoal-pred (car subgoals)) propositional-list))
		    (set! nr-propositionals (+ nr-propositionals 1))))))
	(if (or (< (predsym-arity (subgoal-pred (car subgoals))) min-pred-arity) 
		(and (= min-pred-arity 0) (= nr-propositionals 0)))
	    (set! min-pred-arity (predsym-arity (subgoal-pred (car subgoals)))))
	(if (> (predsym-arity (subgoal-pred (car subgoals))) max-pred-arity)
	    (set! max-pred-arity (predsym-arity (subgoal-pred (car subgoals)))))
	(get-termlist-statistics (subgoal-terms (car subgoals)))
	(get-subgoals-statistics (cdr subgoals)))))

(define (get-termlist-statistics terms)
  (if (pair? terms)
      (begin
	(if (variable (ref (car terms)))
	    (if (memq (ref (car terms)) current-clause-variables)
		(set! current-singleton-variables (remq! (ref (car terms)) current-singleton-variables))
		(begin
		  (set! current-clause-variables (cons (ref (car terms)) current-clause-variables))
		  (set! current-singleton-variables (cons (ref (car terms)) current-singleton-variables))))
	    (begin
	      (if (or (< (functor-arity (car (ref (car terms)))) min-func-arity) 
		      (and (= min-func-arity 0) (= nr-constants 0)))
		  (set! min-func-arity (functor-arity (car (ref (car terms))))))
	      (if (> (functor-arity (car (ref (car terms)))) max-func-arity)
		  (set! max-func-arity (functor-arity (car (ref (car terms))))))
	      (if (constant (ref (car terms)))
		  (if (not (assq (car (ref (car terms))) functor-list))
		      (begin
			(set! functor-list (cons (cons (car (ref (car terms))) 1) functor-list))
			(set! constant-list (cons (cons (car (ref (car terms))) 1) constant-list))
			(set! nr-functors (+ nr-functors 1))
			(set! nr-constants (+ nr-constants 1)))
		      (let ((functor (assq (car (ref (car terms))) functor-list))
			    (constant (assq (car (ref (car terms))) constant-list)))
			(set-cdr! functor (+ (cdr functor) 1))
			(set-cdr! constant (+ (cdr constant) 1))))
		  (begin
		    (if (not (assq (car (ref (car terms))) functor-list))
		      (begin
			(set! functor-list (cons (cons (car (ref (car terms))) 1) functor-list))
			(set! nr-functors (+ nr-functors 1)))
		      (let ((functor (assq (car (ref (car terms))) functor-list)))
			(set-cdr! functor (+ (cdr functor) 1))))
		    (get-termlist-statistics (cdr (ref (car terms))))))))
	(get-termlist-statistics (cdr terms)))))

(define (select-closure-constant)
  (select-constant constant-list #f)
  ;;(if (pair? constant-list)
      ;;(set! selected-constant (list (caar constant-list))))
  (if (or #t dewey-numbers (> debug-level 4))
      (begin
	(if (and selected-constant (not silent))
	    (begin
	      (vector-set! kbo-precedences (- (functor-number (car selected-constant)) 1) 0)
	      (display "constant selected for t-closures: ")
	      (display-ref selected-constant))
	    (display "no constant in input"))
	(newline))))

(define (select-constant constant-list max-constant)
  (if (pair? constant-list)
      (if (or (not max-constant)
	      (> (cdar constant-list) (cdr max-constant)))
	  (select-constant (cdr constant-list) (car constant-list))
	  (select-constant (cdr constant-list) max-constant))
      (if max-constant
	  (set! selected-constant (list (car max-constant))))))

(define (reset-formula-statistics)
  (set! nr-axioms 0)   
  (set! nr-lemmas 0)   
  (set! nr-hypotheses 0)   
  (set! nr-conjectures 0)   
  (set! nr-all-negatives 0)   
  (set! current-clause-variables '())
  (set! current-singleton-variables '())
  (set! predicate-list '())
  (set! functor-list '())
  (set! propositional-list '())
  (set! constant-list '())
  (set! current-positive-subgoals 0)
  (set! nr-clauses 0)
  (set! nr-non-horn 0)
  (set! nr-units 0)
  (set! nr-literals 0)
  (set! nr-eq-literals 0)
  (set! max-clause-length 0)
  (set! nr-predicates 0)
  (set! nr-propositionals 0)
  (set! max-pred-arity 0)
  (set! min-pred-arity 0)
  (set! nr-functors 0)
  (set! nr-constants 0)
  (set! max-func-arity 0)
  (set! min-func-arity 0)
  (set! nr-variables 0)
  (set! nr-singleton-variables 0))

(define (set-all-negatives-relevant clauses)
  (if (pair? clauses)
      (begin
	(if (all-negative (clause-subgoals (car clauses)))
	    (set-clause-relevant (car clauses) #t)
	    (set-clause-relevant (car clauses) #f))
	(set-all-negatives-relevant (cdr clauses)))))

(define (all-negative subgoals)
  (if (null? subgoals)
      #t
      (if (eq? '- (subgoal-sign (car subgoals)))
	  (all-negative (cdr subgoals))
	  #f)))

(define (display-formula-statistics)
  (if preprocess-only
      (display "% ")
      (display "  "))
  (display #"\tSyntax :\t")
  (display #"Number of clauses :\t")
  (display nr-clauses)
  (display " ( ")
  (display nr-non-horn)
  (display " non-Horn; ")
  (display nr-units)
  (display " unit)")
  (newline)
  (if preprocess-only
      (display "% ")
      (display "  "))
  (display #"\t\t\t")
  (display #"Number of literals :\t")
  (display nr-literals)
  (display " ( ")
  (display nr-eq-literals)
  (display " equality)")
  (newline)
  (if preprocess-only
      (display "% ")
      (display "  "))
  (display #"\t\t\t")
  (display #"Maximal clause size :\t")
  (display max-clause-length)
  (newline)
  (if preprocess-only
      (display "% ")
      (display "  "))
  (display #"\t\t\t")
  (display #"Number of predicates :\t")
  (display nr-predicates)
  (display " ( ")
  (display nr-propositionals)
  (display " propositional; ")
  (display min-pred-arity)
  (display "-")
  (display max-pred-arity)
  (display " arity)")
  (newline)
  (if preprocess-only
      (display "% ")
      (display "  "))
  (display #"\t\t\t")
  (display #"Number of functors :\t")
  (display nr-functors)
  (display " ( ")
  (display nr-constants)
  (display " constant; ")
  (display min-func-arity)
  (display "-")
  (display max-func-arity)
  (display " arity)")
  (newline)
  (if preprocess-only
      (display "% ")
      (display "  "))
  (display #"\t\t\t")
  (display #"Number of variables :\t")
  (display nr-variables)
  (display " ( ")
  (display nr-singleton-variables)
  (display " singleton)")
  (newline)
  (if preprocess-only
      (display "% ")
      (display "  "))
  (display #"\t\t\t")
  (display #"Maximal term depth :\t")
  (display input-term-depth)
  (newline))

(define (var-in-list? var list)
  (if (pair? list)
      (if (eq? var (car list))
	  list
	  (var-in-list? var (cdr list)))
      #f))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Demodulation functions
;; ------------------------------------------------------------

(define (is-demodulator eq-unit)
  (or (is-right-demodulator eq-unit) (is-left-demodulator eq-unit)))

;; FIXME: Demodulators are not restricted to complex terms! Change that
(define (is-left-demodulator eq-unit)
  (let ((left (ref (car (subgoal-terms (car (clause-subgoals eq-unit))))))
	(right (ref (cadr (subgoal-terms (car (clause-subgoals eq-unit)))))))
    (and (complex left)
	 (pairwise-different-variable-args (cdr left) '())
	 (not (functor-occurs-in-term (car left) right))
	 (< (simple-term-complexity left) (* demod-limit (simple-term-complexity right))))))

(define (is-right-demodulator eq-unit)
  (let ((left (ref (car (subgoal-terms (car (clause-subgoals eq-unit))))))
	(right (ref (cadr (subgoal-terms (car (clause-subgoals eq-unit)))))))
    (and (complex right)
	 (pairwise-different-variable-args (cdr right) '())
	 (not (functor-occurs-in-term (car right) left))
	 (< (simple-term-complexity left) (* demod-limit (simple-term-complexity right))))))

(define (pairwise-different-variable-args terms arg-vars)
  (if (pair? terms)
      (if (and (variable (ref (car terms))) (not (memq (ref (car terms)) arg-vars)))
	  (pairwise-different-variable-args (cdr terms) (cons (ref (car terms)) arg-vars))
	  #f)
      #t))

(define (functor-occurs-in-term functor term)
  (if (or (index-variable term) (variable term))
      #f
      (if (eq? functor (car term))
	  #t
	  (if (complex term)
	      (functor-occurs-in-termlist functor (cdr term))
	      #f))))

(define (functor-occurs-in-termlist functor terms)
  (if (pair? terms)
      (if (functor-occurs-in-term functor (ref (car terms)))
	  #t
	  (functor-occurs-in-termlist functor (cdr terms)))
      #f))

(define (simple-term-complexity term)
  (if (or (variable term) (constant term)) 
      1
      (+ 1 (simple-termlist-complexity (cdr term)))))

(define (simple-termlist-complexity terms)
  (if (null? terms)
      0
      (+ (simple-term-complexity (car terms)) (simple-termlist-complexity (cdr terms)))))

(define (insert-splitted-clauses-in-clauselist clauses clauselist)
  (if (pair? clauses)
      (insert-new-clause-in-clauselist (car clauses) (insert-splitted-clauses-in-clauselist (cdr clauses) clauselist))
      clauselist))

(define (clear-clauses-path-info clauses)
   (if (pair? clauses)
       (begin
	 (set-clause-path (car clauses) #f)
	 (clear-subgoals-next-info (clause-subgoals (car clauses)))
	 (clear-clauses-path-info (cdr clauses)))))  

(define (clear-subgoals-next-info subgoals)
   (if (pair? subgoals)
       (begin
	 (if (subgoal-next (car subgoals))
	     (begin
	       (set-subgoal-links (car subgoals) (list (subgoal-next (car subgoals))))
	       (set-subgoal-next (car subgoals) #f)))
	 (clear-subgoals-next-info (cdr subgoals)))))  

;;;*************************************************************************
;; ------------------------------------------------------------
;; Functions for preprocess rewriting
;; ------------------------------------------------------------

(define instantiatable-vars '())

(define (match-termlists k l)
  ;; tries to match termlists k and l.
  ;; If matching succeeds, #t is returned, else #f.
  ;; Side effect: termlists and trail destructively modified.
  ;; Before every matching, (trail-stop) must be called in order
  ;; to mark the trail position for the stored substitution
  ;; Also, instantiatable-vars must be set to the list of variables occurring in k
   (if (null? k)
     #t
     (if (match-terms (ref (car k)) (ref (car l)))
	 (match-termlists (cdr k) (cdr l))
	 #f)))

(define (match-terms s t)
  ;; tries to match terms s and t.
  ;; If matching succeeds, #t is returned, else #f.
  ;; Side effect: terms and trail destructively modified.
  ;; Before every unification, (trail-stop) must be called in order
  ;; to mark the trail position for the stored substitution
 (if (eq? s t)
     #t
     (if (variable s)
	 (if (memq s instantiatable-vars)
	     (bind s t)
	     (unbind-vars))
	 (if (variable t)
	     (unbind-vars)
	     (if (eq? (car s) (car t))
	         (match-termlists (cdr s) (cdr t))
                 (unbind-vars))))))

(define (clear-instantiations)
  ;; deletes all instantiations to the previous trailstop, thus making
  ;; those instantiations irreversible
 (if (trail-content)
     (begin (trail-previous) (clear-instantiations))
     (begin (trail-previous) #f)))


(define (try-rewrite-clauses eq-sg clauses)
  (if (pair? clauses)
      (begin
	(if (> debug-level 5)
	    (begin 
	      (display "try rewriting ")
	      (display-clause (car clauses))
	      (newline)))
	(if (not (eq? (car clauses) (subgoal-clause eq-sg)))
	    (begin
	      (if (unit-clause (car clauses))
		  (delete-subgoal-from-index (car (clause-subgoals (car clauses))) 'unit))
	      (set! clause-rewritten #f)
	      (try-rewrite-subgoals eq-sg (clause-subgoals (car clauses)))
	      (if (and clause-rewritten processing-conjecture)
		  (set-input-type (clause-input (car clauses)) 'conjecture))
	      (if (unit-clause (car clauses))
		  (insert-subgoal-in-index (car (clause-subgoals (car clauses))) 'unit))))
	;;(if clause-rewritten
	    ;;(try-rewrite-clauses eq-sg clauses)
	    (try-rewrite-clauses eq-sg (cdr clauses)))))

(define (try-rewrite-subgoals eq-sg subgoals)
  (if (pair? subgoals)
      (begin
	(try-rewrite-terms eq-sg (subgoal-terms (car subgoals)))
	(if (eq-subgoal? (car subgoals)) (check-orientation (car subgoals)))
	(try-rewrite-subgoals eq-sg (cdr subgoals)))))

(define (try-rewrite-terms eq-sg terms)
  (if (pair? terms)
      (begin
	(set! term-rewritten #f)
	(if (not (variable (ref (car terms))))
	    (if (subgoal-ordered-eq eq-sg)
		(try-ordered-rewrite-term eq-sg terms)
		(begin
		  (try-left-rewrite-term eq-sg terms)
		  (try-right-rewrite-term eq-sg terms))))
	(if term-rewritten
	    (try-rewrite-terms eq-sg terms)
	    (try-rewrite-terms eq-sg (cdr terms))))))

(define (try-ordered-rewrite-terms eq-sg terms)
  (if (pair? terms)
      (begin
	(if (not (variable (ref (car terms))))
	    (try-ordered-rewrite-term eq-sg terms))
	(try-ordered-rewrite-terms eq-sg (cdr terms)))))

(define (try-left-rewrite-terms eq-sg terms)
  (if (pair? terms)
      (begin
	(if (not (variable (ref (car terms))))
	    (try-left-rewrite-term eq-sg terms))
	(try-left-rewrite-terms eq-sg (cdr terms)))))

(define (try-right-rewrite-terms eq-sg terms)
  (if (pair? terms)
      (begin
	(if (not (variable (ref (car terms))))
	    (try-right-rewrite-term eq-sg terms))
	(try-right-rewrite-terms eq-sg (cdr terms)))))

(define (try-ordered-rewrite-term eq-sg terms)
  (set! instantiatable-vars (clause-variables (subgoal-clause eq-sg)))
  (if (match-terms-stop (ref (car (subgoal-terms eq-sg))) (ref (car terms)))
      (begin
	(if dewey-numbers
	    (begin
	      (display "using equation ")
	      (display-subgoal eq-sg)
	      (display " for rewriting ")
	      (display-ref (ref (car terms)))))
	(set-car! terms (unrenamed-term-copy (ref (cadr (subgoal-terms eq-sg)))))
	(if dewey-numbers
	    (begin
	      (display " to ")
	      (display-ref (ref (car terms)))
	      (newline)))
	(unbind-vars)
	(set! total-rewrites (+ total-rewrites 1))
	(set! clause-rewritten #t)
	(set! term-rewritten #t)
	(set! formula-changed #t))
      (if (complex (ref (car terms)))
	  (try-ordered-rewrite-terms eq-sg (cdr (ref (car terms)))))))

(define (try-left-rewrite-term eq-sg terms)
  (set! instantiatable-vars (clause-variables (subgoal-clause eq-sg)))
  (if (match-terms-stop (ref (car (subgoal-terms eq-sg))) (ref (car terms)))
      (if (kbo> (ref (car (subgoal-terms eq-sg))) (ref (cadr (subgoal-terms eq-sg))))
	  (begin
	    (if dewey-numbers
		(begin
		  (display "using equation ")
		  (display-subgoal eq-sg)
		  (display " for rewriting ")
		  (display-ref (ref (car terms)))))
	    (set-car! terms (unrenamed-term-copy (ref (cadr (subgoal-terms eq-sg)))))
	    (if dewey-numbers
		(begin
		  (display " to ")
		  (display-ref (ref (car terms)))
		  (newline)))
	    (unbind-vars)
	    (set! total-rewrites (+ total-rewrites 1))
	    (set! clause-rewritten #t)
	    (set! term-rewritten #t)
	    (set! formula-changed #t))
	  (begin
	    (unbind-vars)
	    (if (complex (ref (car terms)))
		(try-left-rewrite-terms eq-sg (cdr (ref (car terms)))))))
      (begin
	(if (complex (ref (car terms)))
	    (try-left-rewrite-terms eq-sg (cdr (ref (car terms))))))))

(define (try-right-rewrite-term eq-sg terms)
  (set! instantiatable-vars (clause-variables (subgoal-clause eq-sg)))
  (if (match-terms-stop (ref (cadr (subgoal-terms eq-sg))) (ref (car terms)))
      (if (kbo> (ref (cadr (subgoal-terms eq-sg))) (ref (car (subgoal-terms eq-sg))))
	  (begin
	    (if dewey-numbers
		(begin
		  (display "using equation ")
		  (display-subgoal eq-sg)
		  (display " for rewriting ")
		  (display-ref (ref (car terms)))))
	    (set-car! terms (unrenamed-term-copy (ref (car (subgoal-terms eq-sg)))))
	    (if dewey-numbers
		(begin
		  (display " to ")
		  (display-ref (ref (car terms)))
		  (newline)))
	    (unbind-vars)
	    (set! total-rewrites (+ total-rewrites 1))
	    (set! clause-rewritten #t)
	    (set! term-rewritten #t)
	    (set! formula-changed #t))
	  (begin
	    (unbind-vars)
	    (if (complex (ref (car terms)))
		(try-right-rewrite-terms eq-sg (cdr (ref (car terms)))))))
      (begin
	(if (complex (ref (car terms)))
	    (try-right-rewrite-terms eq-sg (cdr (ref (car terms))))))))

(define (unrenamed-term-copy term)
  (if (variable term)
      term
      (if (constant term)
	  (cons (car term) '())
	  (cons (car term) (unrenamed-termlist-copy (cdr term))))))

(define (unrenamed-termlist-copy terms)
  (if (pair? terms)
      (cons (unrenamed-term-copy (ref (car terms))) (unrenamed-termlist-copy (cdr terms)))
      '()))

