(module select
	(include "dctp.sch")
	(export linked-terms-ground initial-select)
	(export (select-subgoal subgoals max-subgoal) (compute-subgoal-weights subgoals) (term-symbol-count term depth)
		(termlist-symbol-count terms depth) (compute-link-complexity sg predsg) (current-subgoal-weight subgoal)
		(subgoal-> sga sgb))
	(import (globals "globals.scm") (unification "unification.scm") (subgoal-simplifiable pruning "pruning.scm")
		(display "display.scm")	(ground-closed-with-index index "index.scm")))

(define linked-terms-ground #t)
(define initial-select #t)
(define nr-selections 0)
(define sg-alternation-rate 5)

; (define (term-instantiation term)
; ; computes instantiation value of a term as follows:
; ;           |  0                          if t is a new variable
; ; t-i(t) =  {  1                          if t is a constant or a non-new variable
; ;           |  1/2 + sum(t-i(a_i)/2*n)    if t = f(a_1,...,a_n)
; ;
;  (cond ((not term) 500)
;        ((variable term) (bind term #f) 0)
;        ((constant term) 1000)
;        (else (+ 500 (quotient (termlist-instantiation (cdr term)) 
;                               (* 2 (functor-arity (car term))))))))

; (define (termlist-instantiation termlist)
;  (if (null? (cdr termlist))
;      (term-instantiation (ref-ext (car termlist)))
;      (+ (term-instantiation (ref-ext (car termlist)))
;         (termlist-instantiation (cdr termlist)))))

(define (term-instantiation term depth)
; computes instantiation value of a term as follows:
;           |  0                          if t is a new variable
; t-i(t) =  {  1                          if t is a constant or a non-new variable
;           |  1/2 + sum(t-i(a_i)/2*n)    if t = f(a_1,...,a_n)
;
 (cond ((not term) (quotient 300 depth))
       ((variable term) (bind term #f) 0)
       ((constant term) 1000)
       (else (+ 500 (quotient (termlist-instantiation (cdr term) (+ depth 1)) 2)))))

(define (termlist-instantiation termlist depth)
 (if (null? (cdr termlist))
     (term-instantiation (ref-ext (car termlist)) depth)
     (min (term-instantiation (ref-ext (car termlist)) depth)
	  (termlist-instantiation (cdr termlist) depth))))

(define (ref-ext ext-term)
;; if term is a <bound_variable>, its binding is returned, 
;; otherwise term is returned
 (if (or (not ext-term) (non-variable ext-term) (free-variable ext-term))
     ext-term
     (ref-ext (car ext-term))))

(define (subgoal-instantiation subgoal)
  (if (null? (subgoal-terms subgoal))
      1000
      (termlist-instantiation (subgoal-terms subgoal) 1)))
      ;;(termlist-instantiation (subgoal-terms subgoal))))

(define (compute-subgoal-weights subgoals)
  (if (pair? subgoals)
      (if complexity-measure
	  (begin
	    (set-subgoal-weight (car subgoals) (termlist-symbol-count (subgoal-terms (car subgoals)) 1))
	    (compute-subgoal-weights (cdr subgoals)))
	  (begin
	    (trail-stop)
	    (set-subgoal-weight (car subgoals) (subgoal-instantiation (car subgoals)))
	    (unbind-vars)
	    (compute-subgoal-weights (cdr subgoals))))))
	  
(define (current-subgoal-weight subgoal)
  (if complexity-measure
      (termlist-symbol-count (subgoal-terms subgoal) 1)
      (let (weight)
	(trail-stop)
	(set! weight (subgoal-instantiation subgoal))
 	(unbind-vars)
; 	(display "<")
; 	(display-subgoal subgoal)
; 	(display " ")
; 	(display weight)
; 	(display "> ")
	weight)))
	  
; (define (select-subgoal subgoals max-subgoal)
;   (if (pair? subgoals)
;       (if (subgoal-selected (car subgoals)) 
; 	  (select-subgoal (cdr subgoals) max-subgoal)
; 	  (if (subgoal-next (car subgoals))
; 	      (begin
; 		(set-subgoal-selected (car subgoals) #t)
; 		(car subgoals))
; 	      (if (or (eq? falsum (car subgoals)) (subgoal-simplifiable (car subgoals)))
; 		  (begin
; 		    (set-subgoal-selected (car subgoals) #t)
; 		    (car subgoals))
; 		  (if (or (not max-subgoal) 
; 			  ;;(and (not min-sg-select) (> (subgoal-weight (car subgoals)) (subgoal-weight max-subgoal)))
; 			  ;;(and min-sg-select (< (subgoal-weight (car subgoals)) (subgoal-weight max-subgoal))))
; 			  (and (not min-sg-select) (subgoal-> (car subgoals) max-subgoal))
; 			  (and min-sg-select (subgoal-< (car subgoals) max-subgoal)))
; 		      (select-subgoal (cdr subgoals) (car subgoals))
; 		      (select-subgoal (cdr subgoals) max-subgoal)))))
;       (if max-subgoal
; 	  (begin
; 	    (set-subgoal-selected max-subgoal #t)
; 	    max-subgoal)
; 	  #f)))

(define (select-subgoal subgoals max-subgoal)
  (if (pair? subgoals)
      (let ((simplifiable (simplifiable-or-falsum subgoals)))
	(if simplifiable
	    simplifiable
	    (let ((unselected-next (unselected-with-next subgoals)))
	      (if unselected-next
		  unselected-next
		  (let (sg (open-sg (not-ground-closed subgoals)))
		    (if open-sg
			(set! sg open-sg)
			(set! sg (select-normal-subgoal subgoals max-subgoal)))
		    sg)))))
      #f))

(define (select-normal-subgoal subgoals max-subgoal)
  (if (pair? subgoals)
      (if (subgoal-selected (car subgoals)) 
	  (select-normal-subgoal (cdr subgoals) max-subgoal)
	  (if (or (not max-subgoal) 
		  (and (not min-sg-select) (subgoal-> (car subgoals) max-subgoal))
		  (and min-sg-select (subgoal-< (car subgoals) max-subgoal)))
	      (select-normal-subgoal (cdr subgoals) (car subgoals))
	      (select-normal-subgoal (cdr subgoals) max-subgoal)))
      (if max-subgoal
	  (begin
	    (set-subgoal-selected max-subgoal #t)
	    max-subgoal)
	  #f)))
      
(define (simplifiable-or-falsum subgoals)  
  (if (pair? subgoals)
      (if (subgoal-selected (car subgoals))
	  (simplifiable-or-falsum (cdr subgoals))
	  (if (or (eq? falsum (car subgoals)) (subgoal-simplifiable (car subgoals)))
	      (begin
		(set-subgoal-selected (car subgoals) #t)
		(car subgoals))))
      #f))

(define (unselected-with-next subgoals)
  (if (pair? subgoals)
      (if (and (subgoal-next (car subgoals)) (not (subgoal-selected (car subgoals))))
	  (begin
	    (set-subgoal-selected (car subgoals) #t)
	    (car subgoals))
	  (unselected-with-next (cdr subgoals)))
      #f))

(define (not-ground-closed subgoals)
  (if (pair? subgoals)
      (if (and (not (subgoal-selected (car subgoals))) (ground-closed-with-index (car subgoals)))
	  (begin
	    (set-subgoal-selected (car subgoals) #t)
	    (car subgoals))
	  (not-ground-closed (cdr subgoals)))
      #f))

(define (initial-select-subgoal subgoals max-subgoal)
  (if (pair? subgoals)
      (if (or (subgoal-selected (car subgoals)) (subgoal-simplifiable (car subgoals)))
	  (initial-select-subgoal (cdr subgoals) max-subgoal)
	  (if (or (not max-subgoal) 
		  ;;(and (not min-sg-select) (> (subgoal-weight (car subgoals)) (subgoal-weight max-subgoal)))
		  ;;(and min-sg-select (< (subgoal-weight (car subgoals)) (subgoal-weight max-subgoal))))
		  (and (not min-sg-select) (subgoal-> (car subgoals) max-subgoal))
		  (and min-sg-select (subgoal-< (car subgoals) max-subgoal)))
	      (initial-select-subgoal (cdr subgoals) (car subgoals))
	      (initial-select-subgoal (cdr subgoals) max-subgoal)))
      (if max-subgoal
	  (begin
	    (set-subgoal-selected max-subgoal #t)
	    max-subgoal)
	  #f)))

(define (subgoal-> sga sgb)
  (if (eq? (subgoal-sign sga) (subgoal-sign sgb))
      (if (or (and (eq? (subgoal-sign sga) '+) (eq-subgoal? sga) (not (eq-subgoal? sgb)))
	      (and (eq? (subgoal-sign sga) '-) (not (eq-subgoal? sga)) (eq-subgoal? sgb)))
	  #f  ;; equations less than positive non-eqs, disequations bigger than negative others
	  (if (or (and (eq? (subgoal-sign sga) '+) (not (eq-subgoal? sga)) (eq-subgoal? sgb))
		  (and (eq? (subgoal-sign sga) '-) (eq-subgoal? sga) (not (eq-subgoal? sgb))))
	      #t ;; same as above
	      (if (= (subgoal-weight sga) (subgoal-weight sgb))
		  (if #f ;;(terms-earlier-instantiated (subgoal-terms sga) (subgoal-terms sgb))
		      #t
		      (if #f ;;(terms-earlier-instantiated (subgoal-terms sgb) (subgoal-terms sga))
			  #f
			  (if (= (predsym-weight (subgoal-pred sga)) (predsym-weight (subgoal-pred sgb)))
			      (< (predsym-arity (subgoal-pred sga)) (predsym-arity (subgoal-pred sgb)))
			      (< (predsym-weight (subgoal-pred sga)) (predsym-weight (subgoal-pred sgb))))))
		  (> (subgoal-weight sga) (subgoal-weight sgb)))))
      (if neg-first
	  (if (eq? (subgoal-sign sga) '-)
	      #t
	      #f)
	  (if (eq? (subgoal-sign sga) '+)
	      (if (eq-subgoal? sga)
		  #f
		  #t)
	      (if (eq-subgoal? sgb)
		  #t
		  #f)))))
			  
(define (subgoal-< sga sgb)
  (if (eq? (subgoal-sign sga) (subgoal-sign sgb))
      (if (or (and (eq? (subgoal-sign sga) '+) (eq-subgoal? sga) (not (eq-subgoal? sgb)))
	      (and (eq? (subgoal-sign sga) '-) (not (eq-subgoal? sga)) (eq-subgoal? sgb)))
	  #f  ;; equations less than positive non-eqs, disequations bigger than negative others
	  (if (or (and (eq? (subgoal-sign sga) '+) (not (eq-subgoal? sga)) (eq-subgoal? sgb))
		  (and (eq? (subgoal-sign sga) '-) (eq-subgoal? sga) (not (eq-subgoal? sgb))))
	      #t ;; same as above
	      (if (= (subgoal-weight sga) (subgoal-weight sgb))
		  (if (terms-earlier-instantiated (subgoal-terms sgb) (subgoal-terms sga))
		      #t
		      (if (terms-earlier-instantiated (subgoal-terms sga) (subgoal-terms sgb))
			  #f
			  (if (= (predsym-weight (subgoal-pred sga)) (predsym-weight (subgoal-pred sgb)))
			      (> (predsym-arity (subgoal-pred sga)) (predsym-arity (subgoal-pred sgb)))
			      (> (predsym-weight (subgoal-pred sga)) (predsym-weight (subgoal-pred sgb))))))
		  (< (subgoal-weight sga) (subgoal-weight sgb)))))
      (if neg-first
	  (if (eq? (subgoal-sign sga) '-)
	      #t
	      #f)
	  (if (eq? (subgoal-sign sga) '+)
	      (if (eq-subgoal? sga)
		  #f
		  #t)
	      (if (eq-subgoal? sgb)
		  #t
		  #f)))))
			  
; (define (subgoal-> sga sgb)
;   (if (= (predsym-weight (subgoal-pred sga)) (predsym-weight (subgoal-pred sgb)))
;       (if (= (subgoal-weight sga) (subgoal-weight sgb))
; 	  (if (eq? (subgoal-sign sga) (subgoal-sign sgb))
; 	      (if (= (predsym-arity (subgoal-pred sga)) (predsym-arity (subgoal-pred sgb))) 
; 		  (terms-earlier-instantiated (subgoal-terms sga) (subgoal-terms sgb))
; 		  (> (predsym-arity (subgoal-pred sga)) (predsym-arity (subgoal-pred sgb))))
; 	      (if (eq? (subgoal-sign sga) '-)
; 		  #t
; 		  #f))
; 	  (> (subgoal-weight sga) (subgoal-weight sgb)))
;       (< (predsym-weight (subgoal-pred sga)) (predsym-weight (subgoal-pred sgb)))))

; (define (subgoal-< sga sgb)
;   (if (= (predsym-weight (subgoal-pred sga)) (predsym-weight (subgoal-pred sgb)))
;       (if (= (subgoal-weight sga) (subgoal-weight sgb))
; 	  (if (eq? (subgoal-sign sga) (subgoal-sign sgb))
; 	      (if (= (predsym-arity (subgoal-pred sga)) (predsym-arity (subgoal-pred sgb))) 
; 		  (terms-earlier-instantiated (subgoal-terms sgb) (subgoal-terms sga))
; 		  (< (predsym-arity (subgoal-pred sga)) (predsym-arity (subgoal-pred sgb))))
; 	      (if (eq? (subgoal-sign sga) '+)
; 		  #t
; 		  #f))
; 	  (< (subgoal-weight sga) (subgoal-weight sgb)))
;       (> (predsym-weight (subgoal-pred sga)) (predsym-weight (subgoal-pred sgb)))))

(define (terms-earlier-instantiated terms-a terms-b)
  (if (and (pair? terms-a) (pair? terms-b))
      (let ((result (term-more-instantiated (ref (car terms-a)) (ref (car terms-b)))))
	(if (eq? result 'eq-instantiated)
	    (terms-earlier-instantiated (cdr terms-a) (cdr terms-b))
	    result))
      (pair? terms-b)))

(define (term-more-instantiated k l)
  (if (variable k)
      (if (variable l)
	  'eq-instantiated
	  #f)
      (if (variable l)
	  #t
	  (if (= (functor-weight (car k)) (functor-weight (car l)))
	      (if (= (functor-arity (car k)) (functor-arity (car l)))
		  (if (constant k)
		      'eq-instantiated
		      (terms-earlier-instantiated (cdr k) (cdr l)))
		  (if (< (functor-arity (car k)) (functor-arity (car l)))
		      #t
		      #f))
	      (< (functor-weight (car k)) (functor-weight (car l)))))))


(define (termlist-symbol-count terms depth)
  (if (pair? terms)
      (if maximal-complexity
	  (max (term-symbol-count (ref (car terms)) depth) (termlist-symbol-count (cdr terms) depth))
	  (+ (term-symbol-count (ref (car terms)) depth) (termlist-symbol-count (cdr terms) depth)))
      (begin
	(if (> current-depth max-term-depth)
	    (set! max-term-depth current-depth))
	0)))

(define (term-symbol-count term depth)
  (if (> depth current-depth)
      (set! current-depth depth))
  (if (variable term)
      (begin
	(set! complexity (+ complexity 1))
	(set! linked-terms-ground #f)
	1)
      (if (constant term)
	  (begin
	    (set! complexity (+ complexity 1))
	    0)
	  (begin
	    (set! complexity (+ complexity 1))
	    (+ (* 2 depth depth) (termlist-symbol-count (cdr term) (+ depth 1)))))))

(define (compute-link-complexity sg predsg)
  (+ (compute-link-clause-complexity (clause-subgoals (subgoal-clause sg)) sg)
     (compute-link-clause-complexity (clause-subgoals (subgoal-clause predsg)) predsg)))

(define (compute-link-clause-complexity subgoals linksg)
  (if (pair? subgoals)
      (if (or (subgoal-simplified (car subgoals)) (eq? (car subgoals) linksg))
          (compute-link-clause-complexity (cdr subgoals) linksg)
	  (+ (bit-lsh (termlist-symbol-count (subgoal-terms (car subgoals)) 1) (predsym-weight (subgoal-pred (car subgoals))))
	     (compute-link-clause-complexity (cdr subgoals) linksg)))
      0))

; (define (compute-link-clause-complexity subgoals linksg)
;   (if (pair? subgoals)
;       (if (or (subgoal-simplified (car subgoals)) (eq? (car subgoals) linksg))
; 	  (compute-link-clause-complexity (cdr subgoals) linksg)
; 	  (+ (bit-lsh (termlist-symbol-count (subgoal-terms (car subgoals))) (predsym-weight (subgoal-pred (car subgoals))))
; 	     (compute-link-clause-complexity (cdr subgoals) linksg)))
;       0))
