(module splitting
	(include "dctp.sch")
	(export (split-subgoals subgoals split-lists) (splitlit subgoal) (all-ground-are-splitlits split-lists)
		(move-special-split-list-to-front subgoal split-lists) (move-special-split-list-to-rear subgoal split-lists)
		(split-lists-to-clauses split-lists depth split-lit) (one-list-is-ground split-lists))
	(import (globals "globals.scm") (display "display.scm")  (path "path.scm") (nr-initial-predicates parser "parser.scm")
		(free-vars-in-termlist dctp "dctp.scm") (sign-predsym> subsumption "subsumption.scm")))

(define splitlit-counter 1)
(define split-prefix-string "_splitlit_")
(define split-prefix (string->symbol split-prefix-string))

;; NOTA BENE: The sequence of the subgoals and the way they are consed together does matter indeed
;; as the defined partial order on the subgoals must be preserved!

(define (split-subgoals subgoals split-lists)
  (if (pair? subgoals)
      (let (matching-lists (sg-vars (free-vars-in-termlist (subgoal-terms (car subgoals)) '())))
	(set! matching-lists (find-matching-split-lists sg-vars split-lists))
	(if (= (length matching-lists) 0)
	    (let (new-split-list)
	      (set! new-split-list (list sg-vars sg-vars (list (car subgoals))))
	      (split-subgoals (cdr subgoals) (cons new-split-list split-lists)))
	    (if (= (length matching-lists) 1)
		(begin
		  (add-subgoal-to-split-list (car subgoals) sg-vars (car matching-lists))
		  (split-subgoals (cdr subgoals) split-lists))
		(begin
		  ;;(display-split-lists matching-lists)
		  (set! split-lists (list-subtract split-lists matching-lists))
		  (set! split-lists (cons (join-split-lists matching-lists) split-lists))
		  (add-subgoal-to-split-list (car subgoals) sg-vars (car split-lists))
		  (split-subgoals (cdr subgoals) split-lists)))))
      split-lists))

(define (find-matching-split-lists vars split-lists)
  (if (pair? split-lists)
      (if (pair? vars)
	  (if (nonempty-intersection vars (car (car split-lists)))
	      (cons (car split-lists) (find-matching-split-lists vars (cdr split-lists)))
	      (find-matching-split-lists vars (cdr split-lists)))
	  (if (null? (car (car split-lists)))
	      (cons (car split-lists) (find-matching-split-lists vars (cdr split-lists)))
	      (find-matching-split-lists vars (cdr split-lists))))
      '()))

(define (nonempty-intersection k l)
  (if (pair? k)
      (if (memq (car k) l)
	  (car k)
	  (nonempty-intersection (cdr k) l))
      #f))

(define (add-subgoal-to-split-list subgoal sg-vars split-list)
  ;; add the subgoal vars to the list of split-class vars
  (set-car! split-list (join-lists (car split-list) sg-vars))
  ;; test if there remain any vars common to all subgoals of the class
  (set-car! (cdr split-list) (intersect-lists (cadr split-list) sg-vars))
  ;; add the subgoal to the list of subgoals
  (set-car! (cddr split-list) (append! (caddr split-list) (list subgoal))))

(define (join-lists k l)
  (if (pair? l)
      (if (memq (car l) k)
	  (join-lists k (cdr l))
	  (cons (car l) (join-lists k (cdr l))))
      k))

(define (intersect-lists k l)
  (if (pair? l)
      (if (memq (car l) k)
	  (cons (car l) (intersect-lists k (cdr l)))
	  (intersect-lists k (cdr l)))
      '()))

(define (list-subtract k l)
  (if (pair? l)
      (remq! (car l) (list-subtract k (cdr l)))
      k))

(define (join-split-lists l)
  (if (pair? l)
      (if (pair? (cdr l))
	  (join-split-list (car l) (join-split-lists (cdr l)))
	  (car l))
      (list '() '() '())))

(define (join-split-list k l)
  (set-car! k (join-lists (car k) (car l)))
  (set-car! (cdr k) (intersect-lists (cadr k) (cadr l)))
  (set-car! (cddr k) (append! (caddr k) (caddr l)))
  k)

(define (display-split-lists split-lists)
  (if (pair? split-lists)
      (begin
	(display "join<")
	(display-ref-list (car (car split-lists)))
	(display "> meet<")
	(display-ref-list (cadr (car split-lists)))
	(display "> sg<")
	(display-subgoal-list (caddr (car split-lists)))
	(display ">")
	(newline)
	(display-split-lists (cdr split-lists)))))
      
(define (new-splitlit-pair variable)
  (let ((split-pred (new-predsym)) (pos-split-sg (new-subgoal)) (neg-split-sg (new-subgoal)))
    (set-predsym-symbol split-pred (gensym split-prefix))
    (set-predsym-number split-pred (+ splitlit-counter nr-initial-predicates))
    (if (>= (predsym-number split-pred) path-vector-size)
	(enlarge-path-literal-vectors))
    (set-predsym-weight split-pred 1)
    (set-predsym-pos-context-index split-pred (cons 0 '()))
    (set-predsym-neg-context-index split-pred (cons 0 '()))
    (set-predsym-pos-gr-index split-pred (cons 0 '()))
    (set-predsym-neg-gr-index split-pred (cons 0 '()))
    (set-predsym-pos-unit-index split-pred (cons 0 '()))
    (set-predsym-neg-unit-index split-pred (cons 0 '()))
    (set-predsym-pos-reg-index split-pred (cons 0 '()))
    (set-predsym-neg-reg-index split-pred (cons 0 '()))
    (set! splitlit-counter (+ splitlit-counter 1))
    (if variable
	(begin
	  (set-predsym-arity split-pred 1)
	  (set-subgoal-terms pos-split-sg (list variable))
	  (set-subgoal-terms neg-split-sg (list variable)))
	(begin
	  (set-predsym-arity split-pred 0)
	  (set-subgoal-terms pos-split-sg '())
	  (set-subgoal-terms neg-split-sg '())))
    (set-subgoal-sign pos-split-sg '+)
    (set-subgoal-sign neg-split-sg '-)
    (set-subgoal-pred pos-split-sg split-pred)
    (set-subgoal-pred neg-split-sg split-pred)
    (if (not complexity-measure)
	(begin
	    (set-subgoal-weight pos-split-sg 0)
	    (set-subgoal-weight neg-split-sg 0))
	(begin
	    (set-subgoal-weight pos-split-sg 1000)
	    (set-subgoal-weight neg-split-sg 1000)))
    (set-subgoal-links pos-split-sg '())
    (set-subgoal-links neg-split-sg '())
    (set-subgoal-next pos-split-sg neg-split-sg)
    (set-subgoal-next neg-split-sg pos-split-sg)
    (cons pos-split-sg neg-split-sg)))

;; the following functions do not work as planned if subgoal is not found in a splitlist
(define (move-special-split-list-to-front subgoal split-lists)
  (if (and subgoal (pair? split-lists))
      (if (memq subgoal (caddr (car split-lists)))
	  split-lists
	  (let ((tmp (find-special-split-list subgoal (cdr split-lists))))
	    (if tmp
		(cons tmp (remq! tmp split-lists))
		split-lists)))
      split-lists))

(define (find-special-split-list subgoal split-lists)
  (if (pair? split-lists)
      (if (memq subgoal (caddr (car split-lists)))
	  (car split-lists)
	  (find-special-split-list subgoal (cdr split-lists)))
      #f))

(define (move-special-split-list-to-rear subgoal split-lists)
  (if (and subgoal (pair? split-lists))
      (if (pair? (cdr split-lists))
	  (extract-and-append-split-list subgoal split-lists)
	  split-lists)
      split-lists))

(define (extract-and-append-split-list subgoal split-lists)
  (if (pair? split-lists)
      (if (memq subgoal (caddr (car split-lists)))
	  (if (pair? (cdr split-lists))
	      (let ((tmp (car split-lists)))
		(set-car! split-lists (cadr split-lists))
		(set-cdr! split-lists (cddr split-lists))
		(append! split-lists (cons tmp '()))
		split-lists)
	      split-lists)
	  (begin
	    (extract-and-append-split-list subgoal (cdr split-lists))
	    split-lists))
      split-lists))

(define (insert-splitlit-in-subgoals subgoals split-lit)
  (if (pair? subgoals)
      (begin
	(if (sign-predsym> (car subgoals) split-lit)
	    (let ((tmp (cons (car subgoals) (cdr subgoals))))
	      (set-car! subgoals split-lit)
	      (set-cdr! subgoals tmp))
	    (if (pair? (cdr subgoals))
		(insert-splitlit-in-subgoals (cdr subgoals) split-lit)
		(set-cdr! subgoals (cons split-lit '()))))
	subgoals)
      (cons split-lit '())))
	      
(define (split-lists-to-clauses split-lists depth split-lit)
  (if (pair? split-lists)
      (let ((cl (new-clause)))
	(set-clause-foldup-lits cl '())
	(if preprocessing
	    (set-clause-initial cl #t))
	(set-clause-variables cl (caar split-lists))
	(set-clause-nr-variables cl (length (caar split-lists)))
	(if split-lit
	    (begin
	      (set-clause-subgoals cl (insert-splitlit-in-subgoals (caddar split-lists) split-lit))
	      (if (not preprocessing)
		  (begin
		    (set-clause-path cl (subgoal-next split-lit))
		    (set-subgoal-next split-lit #f))))
	    (set-clause-subgoals cl (caddar split-lists)))
	(set-clause-depth cl depth)
	(set-clause-unsimplified cl (length (clause-subgoals cl)))
	(if (pair? (cdr split-lists))
	    (let ((new-split-lits (new-splitlit-pair #f)) tmp-result)
	      (set-clause-subgoals cl (insert-splitlit-in-subgoals (clause-subgoals cl) (car new-split-lits)))
	      (set-clause-unsimplified cl (length (clause-subgoals cl)))
	      (set! tmp-result (split-lists-to-clauses (cdr split-lists) (+ depth 1) (cdr new-split-lits)))
	      (if preprocessing
		  (begin
		    (set-subgoal-links (car new-split-lits) (list (cdr new-split-lits)))
		    (set-subgoal-links (cdr new-split-lits) (list (car new-split-lits)))
		    (set-subgoal-next (car new-split-lits) #f)
		    (set-subgoal-next (cdr new-split-lits) #f))
		  (set-subgoal-next (car new-split-lits) (car tmp-result)))
	      (cons cl tmp-result))
	    (cons cl '())))))

(define (splitlit subgoal)
  (substring=? split-prefix-string (symbol->string (predsym-symbol (subgoal-pred subgoal))) 10))

(define (all-ground-are-splitlits split-lists)
  (if (pair? split-lists)
      (if (pair? (caar split-lists))
	  (all-ground-are-splitlits (cdr split-lists))
	  (all-subgoals-are-splitlits (caddar split-lists)))
      #t))

(define (all-subgoals-are-splitlits subgoals)
  (if (pair? subgoals)
      (if (splitlit (car subgoals))
	  (all-subgoals-are-splitlits (cdr subgoals))
	  #f)
      #t))
(define (one-list-is-ground split-lists)
  (if (pair? split-lists)
      (if (pair? (caar split-lists))
	  (one-list-is-ground (cdr split-lists))
	  #t)
      #f))